# Copyright Universitat Pompeu Fabra 2020-2023  https://www.compscience.org
# Distributed under the MIT License.
# (See accompanying file README.md file or copy at http://opensource.org/licenses/MIT)

import h5py
import numpy as np
import os
import torch as pt
from torch_geometric.data import Data
from torchmdnet.datasets.memdataset import MemmappedDataset
from tqdm import tqdm


class QM9q(MemmappedDataset):
    HARTREE_TO_EV = 27.211386246  #::meta private:
    BORH_TO_ANGSTROM = 0.529177  #::meta private:
    DEBYE_TO_EANG = 0.2081943  #::meta private: Debey -> e*A

    # Ion energies of elements
    ELEMENT_ENERGIES = {
        1: {0: -0.5013312007, 1: 0.0000000000},
        6: {-1: -37.8236383010, 0: -37.8038423252, 1: -37.3826165878},
        7: {-1: -54.4626446440, 0: -54.5269367415, 1: -53.9895574739},
        8: {-1: -74.9699154500, 0: -74.9812632126, 1: -74.4776884006},
        9: {-1: -99.6695561536, 0: -99.6185158728},
    }  #::meta private:

    # Select an ion with the lowest energy for each element
    INITIAL_CHARGES = {
        element: sorted(zip(charges.values(), charges.keys()))[0][1]
        for element, charges in ELEMENT_ENERGIES.items()
    }  #::meta private:

    def __init__(
        self,
        root=None,
        transform=None,
        pre_transform=None,
        pre_filter=None,
        paths=None,
    ):
        self.name = self.__class__.__name__
        self.paths = str(paths)
        super().__init__(
            root,
            transform,
            pre_transform,
            pre_filter,
            properties=("y", "neg_dy", "q", "pq", "dp"),
        )

    @property
    def raw_paths(self):
        paths = self.paths

        if os.path.isfile(paths):
            return [paths]
        if os.path.isdir(paths):
            return [
                os.path.join(paths, file_)
                for file_ in os.listdir(paths)
                if file_.endswith(".h5")
            ]

        raise RuntimeError(f"Cannot load {paths}")

    @staticmethod
    def compute_reference_energy(atomic_numbers, charge):
        atomic_numbers = np.array(atomic_numbers)
        charge = int(charge)

        charges = [QM9q.INITIAL_CHARGES[z] for z in atomic_numbers]
        energy = sum(
            QM9q.ELEMENT_ENERGIES[z][q] for z, q in zip(atomic_numbers, charges)
        )

        while sum(charges) != charge:
            dq = np.sign(charge - sum(charges))

            new_energies = []
            for i, (z, q) in enumerate(zip(atomic_numbers, charges)):
                if (q + dq) in QM9q.ELEMENT_ENERGIES[z]:
                    new_energy = (
                        energy
                        - QM9q.ELEMENT_ENERGIES[z][q]
                        + QM9q.ELEMENT_ENERGIES[z][q + dq]
                    )
                    new_energies.append((new_energy, i, q + dq))

            energy, i, q = sorted(new_energies)[0]
            charges[i] = q

        assert sum(charges) == charge

        energy = sum(
            QM9q.ELEMENT_ENERGIES[z][q] for z, q in zip(atomic_numbers, charges)
        )

        return energy * QM9q.HARTREE_TO_EV

    def sample_iter(self, mol_ids=False):
        for path in tqdm(self.raw_paths, desc="Files"):
            molecules = list(h5py.File(path).values())[0].items()

            for mol_id, mol in tqdm(molecules, desc="Molecules", leave=False):
                z = pt.tensor(mol["atomic_numbers"], dtype=pt.long)

                for conf in mol["energy"]:
                    assert mol["positions"].attrs["units"] == "Å : ångströms"
                    pos = pt.tensor(mol["positions"][conf], dtype=pt.float32)
                    assert z.shape[0] == pos.shape[0]
                    assert pos.shape[1] == 3

                    assert mol["energy"].attrs["units"] == "E_h : hartree"
                    y = (
                        pt.tensor(mol["energy"][conf][()], dtype=pt.float64)
                        * self.HARTREE_TO_EV
                    )

                    assert (
                        mol["gradient_vector"].attrs["units"]
                        == "vector : Hartree/Bohr "
                    )
                    neg_dy = (
                        -pt.tensor(mol["gradient_vector"][conf], dtype=pt.float32)
                        * self.HARTREE_TO_EV
                        / self.BORH_TO_ANGSTROM
                    )
                    assert z.shape[0] == neg_dy.shape[0]
                    assert neg_dy.shape[1] == 3

                    assert (
                        mol["electronic_charge"].attrs["units"]
                        == "n : fractional electrons"
                    )
                    pq = pt.tensor(mol["electronic_charge"][conf], dtype=pt.float32)
                    q = pq.sum().round().to(pt.long)

                    assert mol["dipole_moment"].attrs["units"] == "\\mu : Debye "
                    dp = (
                        pt.tensor(mol["dipole_moment"][conf], dtype=pt.float32)
                        * self.DEBYE_TO_EANG
                    )

                    y -= self.compute_reference_energy(z, q)

                    # Skip samples with large forces
                    if neg_dy.norm(dim=1).max() > 100:  # eV/A
                        continue

                    # Create a sample
                    args = dict(
                        z=z, pos=pos, y=y.view(1, 1), neg_dy=neg_dy, q=q, pq=pq, dp=dp
                    )
                    if mol_ids:
                        args["mol_id"] = mol_id
                    data = Data(**args)

                    if self.pre_filter is not None and not self.pre_filter(data):
                        continue

                    if self.pre_transform is not None:
                        data = self.pre_transform(data)

                    yield data



# Copyright Universitat Pompeu Fabra 2020-2023  https://www.compscience.org
# Distributed under the MIT License.
# (See accompanying file README.md file or copy at http://opensource.org/licenses/MIT)

import os
import os.path as osp
from typing import Callable, List, Optional
import numpy as np
import torch
from torch_geometric.data import (
    Data,
    InMemoryDataset,
    download_url,
    extract_tar,
    extract_zip,
)


class MD22(InMemoryDataset):

    gdml_url = "http://quantum-machine.org/gdml/data/npz"

    file_names = {
        "AT-AT-CG-CG": "md22_AT-AT-CG-CG.npz",
        "AT-AT": "md22_AT-AT.npz",
        "Ac-Ala3-NHMe": "md22_Ac-Ala3-NHMe.npz",
        "DHA": "md22_DHA.npz",
        "buckyball-catcher": "md22_buckyball-catcher.npz",
        "dw-nanotube": "md22_dw_nanotube.npz",
        "stachyose": "md22_stachyose.npz",
    }

    def __init__(
        self,
        root: str,
        molecules: str,
        transform: Optional[Callable] = None,
        pre_transform: Optional[Callable] = None,
        pre_filter: Optional[Callable] = None,
    ):
        name = molecules
        if name not in self.file_names:
            raise ValueError(f"Unknown dataset name '{name}'")

        self.name = name

        super().__init__(root, transform, pre_transform, pre_filter)

        idx = 0
        self.data, self.slices = torch.load(self.processed_paths[idx])

    def mean(self) -> float:
        return float(self._data.energy.mean())

    @property
    def raw_dir(self) -> str:
        return osp.join(self.root, self.name, "raw")

    @property
    def processed_dir(self) -> str:
        return osp.join(self.root, "processed", self.name)

    @property
    def raw_file_names(self) -> str:
        name = self.file_names[self.name]
        return name

    @property
    def processed_file_names(self) -> List[str]:
        return ["data.pt"]

    def download(self):
        url = f"{self.gdml_url}/{self.file_names[self.name]}"
        path = download_url(url, self.raw_dir)

    def process(self):
        it = zip(self.raw_paths, self.processed_paths)
        for raw_path, processed_path in it:
            raw_data = np.load(raw_path)

            z = torch.from_numpy(raw_data["z"]).long()
            pos = torch.from_numpy(raw_data["R"]).float()
            energy = torch.from_numpy(raw_data["E"]).float()
            force = torch.from_numpy(raw_data["F"]).float()

            data_list = []
            for i in range(pos.size(0)):
                data = Data(z=z, pos=pos[i], y=energy[i].unsqueeze(-1), neg_dy=force[i])
                if self.pre_filter is not None and not self.pre_filter(data):
                    continue
                if self.pre_transform is not None:
                    data = self.pre_transform(data)
                data_list.append(data)

            torch.save(self.collate(data_list), processed_path)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({len(self)}, name='{self.name}')"



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""Warp kernels for graph edge index transformation to sparse CSR format."""

import warp as wp


@wp.kernel
def count_row_col(
    edge_index: wp.array(ndim=2, dtype=wp.int32),
    row_count: wp.array(ndim=1, dtype=wp.int32),
    col_count: wp.array(ndim=1, dtype=wp.int32),
):
    tid = wp.tid()

    # skip dummy edges that will exist if static_shapes==True
    if edge_index[0, tid] == -1 or edge_index[1, tid] == -1:
        return

    shift = edge_index.dtype(1)
    wp.atomic_add(row_count, edge_index[0, tid] + shift, wp.int32(1))
    wp.atomic_add(col_count, edge_index[1, tid] + shift, wp.int32(1))


@wp.kernel
def convert_to_sparse(
    edge_index: wp.array(ndim=2, dtype=wp.int32),
    row_count: wp.array(ndim=1, dtype=wp.int32),
    col_count: wp.array(ndim=1, dtype=wp.int32),
    row_indptr: wp.array(ndim=1, dtype=wp.int32),
    col_indptr: wp.array(ndim=1, dtype=wp.int32),
    row_indices: wp.array(ndim=1, dtype=wp.int32),
    col_indices: wp.array(ndim=1, dtype=wp.int32),
    row_data: wp.array(ndim=1, dtype=wp.int32),
    col_data: wp.array(ndim=1, dtype=wp.int32),
):
    tid = wp.tid()
    shift = edge_index.dtype(1)

    src_id = edge_index[0, tid]
    dst_id = edge_index[1, tid]

    # skip dummy edges that will exist if static_shapes==True
    if src_id == -1 or dst_id == -1:
        return

    src_cnt = wp.atomic_sub(row_count, src_id + shift, wp.int32(1))
    dst_cnt = wp.atomic_sub(col_count, dst_id + shift, wp.int32(1))

    row_indices[row_indptr[src_id + shift] - src_cnt] = dst_id
    row_data[row_indptr[src_id + shift] - src_cnt] = wp.int32(tid)

    col_indices[col_indptr[dst_id + shift] - dst_cnt] = src_id
    col_data[col_indptr[dst_id + shift] - dst_cnt] = wp.int32(tid)



# Copyright Universitat Pompeu Fabra 2020-2023  https://www.compscience.org
# Distributed under the MIT License.
# (See accompanying file README.md file or copy at http://opensource.org/licenses/MIT)

import hashlib
import h5py
import numpy as np
import os
import torch as pt
from torchmdnet.datasets.memdataset import MemmappedDataset
from torch_geometric.data import Data, download_url
from tqdm import tqdm
import logging


class SPICE(MemmappedDataset):
    """
    SPICE dataset (https://github.com/openmm/spice-dataset)

    The dataset has several versions (https://github.com/openmm/spice-dataset/releases).
    The version can be selected with `version`. By default, version 1.1.3 is loaded.

    >>> ds = SPICE(".", version="1.1.3")

    The dataset consists of several subsets (https://github.com/openmm/spice-dataset/blob/main/downloader/config.yaml).
    The subsets can be selected with `subsets`. By default, all the subsets are loaded.

    For example, this loads just two subsets:
    >>> ds = SPICE(".", subsets=["SPICE PubChem Set 1 Single Points Dataset v1.2", "SPICE PubChem Set 2 Single Points Dataset v1.2"])

    The loader can filter conformations with large gradients. The maximum gradient norm threshold
    can be set with `max_gradient`. By default, the filter is not applied.

    For example, the filter the threshold is set to 100 eV/A:
    >>> ds = SPICE(".", max_gradient=100)

    The molecules can be subsampled by loading only every `subsample_molecules`-th molecule.
    By default is `subsample_molecules` is set to 1 (load all the molecules).

    For example, only every 10th molecule is loaded:
    >>> ds = SPICE(".", subsample_molecules=10)
    """

    HARTREE_TO_EV = 27.211386246
    BORH_TO_ANGSTROM = 0.529177

    VERSIONS = {
        "1.0": {
            "url": "https://github.com/openmm/spice-dataset/releases/download/1.0",
            "file": "SPICE.hdf5",
        },
        "1.1": {
            "url": "https://github.com/openmm/spice-dataset/releases/download/1.1",
            "file": "SPICE.hdf5",
        },
        "1.1.1": {
            "url": "https://zenodo.org/record/7258940/files",
            "file": "SPICE-1.1.1.hdf5",
            "hash": "5411e7014c6d18ff07d108c9ad820b53",
        },
        "1.1.2": {
            "url": "https://zenodo.org/record/7338495/files",
            "file": "SPICE-1.1.2.hdf5",
            "hash": "a2b5ae2d1f72581040e1cceb20a79a33",
        },
        "1.1.3": {
            "url": "https://zenodo.org/record/7606550/files",
            "file": "SPICE-1.1.3.hdf5",
            "hash": "be93706b3bb2b2e327b690b185905856",
        },
        "1.1.4": {
            "url": "https://zenodo.org/records/8222043/files",
            "file": "SPICE-1.1.4.hdf5",
            "hash": "f27d4c81da0e37d6547276bf6b4ae6a1",
        },
        "2.0.1": {
            "url": "https://zenodo.org/records/10975225/files",
            "file": "SPICE-2.0.1.hdf5",
            "hash": "bfba2224b6540e1390a579569b475510",
        },
    }

    @property
    def raw_dir(self):
        return os.path.join(super().raw_dir, "spice", self.version)

    @property
    def raw_file_names(self):
        return self.VERSIONS[self.version]["file"]

    @property
    def raw_url(self):
        return f"{self.VERSIONS[self.version]['url']}/{self.VERSIONS[self.version]['file']}"

    def __init__(
        self,
        root=None,
        transform=None,
        pre_transform=None,
        pre_filter=None,
        version="1.1.3",
        subsets=None,
        max_gradient=None,
        subsample_molecules=1,
    ):
        arg_hash = f"{version}{subsets}{max_gradient}{subsample_molecules}"
        arg_hash = hashlib.md5(arg_hash.encode()).hexdigest()
        self.name = f"{self.__class__.__name__}-{arg_hash}"
        self.version = str(version)
        assert self.version in self.VERSIONS
        self.subsets = subsets
        self.max_gradient = max_gradient
        self.subsample_molecules = int(subsample_molecules)
        super().__init__(
            root,
            transform,
            pre_transform,
            pre_filter,
            properties=("y", "neg_dy"),
        )

    def sample_iter(self, mol_ids=False):
        assert len(self.raw_paths) == 1
        assert self.subsample_molecules > 0

        molecules = h5py.File(self.raw_paths[0]).items()
        for i_mol, (mol_id, mol) in tqdm(enumerate(molecules), desc="Molecules"):
            if self.subsets:
                if mol["subset"][0].decode() not in list(self.subsets):
                    continue

            # Subsample molecules
            if i_mol % self.subsample_molecules != 0:
                continue

            z = pt.tensor(np.array(mol["atomic_numbers"]), dtype=pt.long)
            all_pos = (
                pt.tensor(np.array(mol["conformations"]), dtype=pt.float32)
                * self.BORH_TO_ANGSTROM
            )
            all_y = (
                pt.tensor(np.array(mol["formation_energy"]), dtype=pt.float64)
                * self.HARTREE_TO_EV
            )
            all_neg_dy = (
                -pt.tensor(np.array(mol["dft_total_gradient"]), dtype=pt.float32)
                * self.HARTREE_TO_EV
                / self.BORH_TO_ANGSTROM
            )
            if all_pos.ndim < 3:
                logging.warning(f"Bogus conformation {mol_id}")
                logging.warning(
                    f"Found {all_pos.shape} positions, {all_y.shape} energies and {all_neg_dy.shape} gradients"
                )
                continue
            assert all_pos.shape[0] == all_y.shape[0]
            assert all_pos.shape[1] == z.shape[0]
            assert all_pos.shape[2] == 3

            assert all_neg_dy.shape[0] == all_y.shape[0]
            assert all_neg_dy.shape[1] == z.shape[0]
            assert all_neg_dy.shape[2] == 3

            for pos, y, neg_dy in zip(all_pos, all_y, all_neg_dy):
                # Skip samples with large forces
                if self.max_gradient:
                    if neg_dy.norm(dim=1).max() > float(self.max_gradient):
                        continue

                # Create a sample
                args = dict(z=z, pos=pos, y=y.view(1, 1), neg_dy=neg_dy)
                if mol_ids:
                    args["mol_id"] = mol_id
                data = Data(**args)

                if self.pre_filter is not None and not self.pre_filter(data):
                    continue

                if self.pre_transform is not None:
                    data = self.pre_transform(data)

                yield data

    def download(self):
        download_url(self.raw_url, self.raw_dir)
        if "hash" in self.VERSIONS[self.version]:
            with open(self.raw_paths[0], "rb") as f:
                file_hash = hashlib.md5(f.read()).hexdigest()
            assert file_hash == self.VERSIONS[self.version]["hash"]



# Copyright Universitat Pompeu Fabra 2020-2023  https://www.compscience.org
# Distributed under the MIT License.
# (See accompanying file README.md file or copy at http://opensource.org/licenses/MIT)

import os
import h5py
import torch
import math
import logging
import numpy as np
from tqdm import tqdm
from os.path import join as opj
from torch_geometric.data import Dataset, Data
import urllib.request
from collections import defaultdict

logger = logging.getLogger("MDCATH")


def load_pdb_list(pdb_list):
    """Load PDB list from a file or return list directly."""
    if isinstance(pdb_list, list):
        return pdb_list
    elif isinstance(pdb_list, str) and os.path.isfile(pdb_list):
        logger.info(f"Reading PDB list from {pdb_list}")
        with open(pdb_list, "r") as file:
            return [line.strip() for line in file]
    raise ValueError("Invalid PDB list. Please provide a list or a path to a file.")


class MDCATH(Dataset):
    def __init__(
        self,
        root,
        transform=None,
        pre_transform=None,
        pre_filter=None,
        source_file="mdcath_source.h5",
        file_basename="mdcath_dataset",
        numAtoms=5000,
        numNoHAtoms=None,
        numResidues=1000,
        temperatures=["348"],
        skip_frames=1,
        pdb_list=None,
        min_gyration_radius=None,
        max_gyration_radius=None,
        alpha_beta_coil=None,
        solid_ss=None,
        numFrames=None,
    ):
        """mdCATH dataset class for PyTorch Geometric to load protein structures and dynamics from the mdCATH dataset.

        Parameters:
        -----------
        root: str
            Root directory where the dataset should be stored. Data will be downloaded to 'root/'.
        numAtoms: int
            Max number of atoms in the protein structure.
        source_file: str
            Name of the source file with the information about the protein structures. Default is "mdcath_source.h5".
        file_basename: str
            Base name of the hdf5 files. Default is "mdcath_dataset".
        numNoHAtoms: int
            Max number of non-hydrogen atoms in the protein structure, not available for original mdcath dataset. Default is None.
            Be sure to have the attribute 'numNoHAtoms' in the source file.
        numResidues: int
            Max number of residues in the protein structure.
        temperatures: list
            List of temperatures (in Kelvin) to download. Default is ["348"]. Available temperatures are ['320', '348', '379', '413', '450']
        skip_frames: int
            Number of frames to skip in the trajectory. Default is 1.
        pdb_list: list or str
            List of PDB IDs to download or path to a file with the PDB IDs. If None, all available PDB IDs from 'source_file' will be downloaded.
            The filters will be applied to the PDB IDs in this list in any case. Default is None.
        min_gyration_radius: float
            Minimum gyration radius (in nm) of the protein structure. Default is None.
        max_gyration_radius: float
            Maximum gyration radius (in nm) of the protein structure. Default is None.
        alpha_beta_coil: tuple
            Tuple with the minimum percentage of alpha-helix, beta-sheet and coil residues in the protein structure. Default is None.
        solid_ss: float
            minimum percentage of solid secondary structure in the protein structure (alpha + beta)/total_residues * 100. Default is None.
        numFrames: int
            Minimum number of frames in the trajectory in order to be considered. Default is None.
        """

        self.url = "https://huggingface.co/datasets/compsciencelab/mdCATH/resolve/main/"
        self.source_file = source_file
        self.file_basename = file_basename
        self.numNoHAtoms = numNoHAtoms
        self.root = root
        os.makedirs(root, exist_ok=True)
        self.numAtoms = numAtoms
        self.numResidues = numResidues
        self.temperatures = [str(temp) for temp in temperatures]
        self.skip_frames = skip_frames
        self.pdb_list = load_pdb_list(pdb_list) if pdb_list is not None else None
        self.min_gyration_radius = min_gyration_radius
        self.max_gyration_radius = max_gyration_radius
        self.alpha_beta_coil = alpha_beta_coil
        self.numFrames = numFrames
        self.solid_ss = solid_ss
        self._ensure_source_file()
        self._filter_and_prepare_data()
        self.idx = None
        super(MDCATH, self).__init__(root, transform, pre_transform, pre_filter)
        # Calculate the total size of the dataset in MB
        self.total_size_mb = self.calculate_dataset_size()

        logger.info(f"Total number of domains: {len(self.processed.keys())}")
        logger.info(f"Total number of conformers: {self.num_conformers}")
        logger.info(f"Total size of dataset: {self.total_size_mb} MB")

    @property
    def raw_file_names(self):
        return [f"{self.file_basename}_{pdb_id}.h5" for pdb_id in self.processed.keys()]

    @property
    def raw_dir(self):
        # Override the raw_dir property to return the root directory
        # The files will be downloaded to the root directory, compatible only with original mdcath dataset
        return self.root

    def _ensure_source_file(self):
        """Ensure the source file is downloaded before processing."""
        source_path = os.path.join(self.root, self.source_file)
        if not os.path.exists(source_path):
            assert (
                self.source_file == "mdcath_source.h5"
            ), "Only 'mdcath_source.h5' is supported as source file for download."
            logger.info(f"Downloading source file {self.source_file}")
            urllib.request.urlretrieve(opj(self.url, self.source_file), source_path)

    def download(self):
        for pdb_id in self.processed.keys():
            file_name = f"{self.file_basename}_{pdb_id}.h5"
            file_path = opj(self.raw_dir, file_name)
            if not os.path.exists(file_path):
                assert (
                    self.file_basename == "mdcath_dataset"
                ), "Only 'mdcath_dataset' is supported as file_basename for download."
                # Download the file if it does not exist
                urllib.request.urlretrieve(opj(self.url, "data", file_name), file_path)

    def calculate_dataset_size(self):
        total_size_bytes = 0
        for pdb_id in self.processed.keys():
            file_name = f"{self.file_basename}_{pdb_id}.h5"
            total_size_bytes += os.path.getsize(opj(self.root, file_name))
        total_size_mb = round(total_size_bytes / (1024 * 1024), 4)
        return total_size_mb

    def _filter_and_prepare_data(self):
        source_info_path = os.path.join(self.root, self.source_file)

        self.processed = defaultdict(list)
        self.num_conformers = 0

        with h5py.File(source_info_path, "r") as file:
            domains = file.keys() if self.pdb_list is None else self.pdb_list

            for pdb_id in tqdm(domains, desc="Processing mdcath source"):
                pdb_group = file[pdb_id]
                if (
                    self.numAtoms is not None
                    and pdb_group.attrs["numProteinAtoms"] > self.numAtoms
                ):
                    continue
                if (
                    self.numResidues is not None
                    and pdb_group.attrs["numResidues"] > self.numResidues
                ):
                    continue
                self._process_temperatures(pdb_id, pdb_group)

    def _process_temperatures(self, pdb_id, pdb_group):
        for temp in self.temperatures:
            for replica in pdb_group[temp].keys():
                self._evaluate_replica(pdb_id, temp, replica, pdb_group)

    def _evaluate_replica(self, pdb_id, temp, replica, pdb_group):
        conditions = [
            self.numFrames is not None
            and pdb_group[temp][replica].attrs["numFrames"] < self.numFrames,
            self.min_gyration_radius is not None
            and pdb_group[temp][replica].attrs["min_gyration_radius"]
            < self.min_gyration_radius,
            self.max_gyration_radius is not None
            and pdb_group[temp][replica].attrs["max_gyration_radius"]
            > self.max_gyration_radius,
            self._evaluate_structure(pdb_group, temp, replica),
            self.numNoHAtoms is not None
            and pdb_group.attrs["numNoHAtoms"] > self.numNoHAtoms,
        ]
        if any(conditions):
            return

        num_frames = math.ceil(
            pdb_group[temp][replica].attrs["numFrames"] / self.skip_frames
        )
        self.processed[pdb_id].append((temp, replica, num_frames))
        self.num_conformers += num_frames

    def _evaluate_structure(self, pdb_group, temp, replica):
        alpha = pdb_group[temp][replica].attrs["alpha"]
        beta = pdb_group[temp][replica].attrs["beta"]
        solid_ss = (alpha + beta) / pdb_group.attrs["numResidues"] * 100
        return self.solid_ss is not None and solid_ss < self.solid_ss

    def len(self):
        return self.num_conformers

    def _setup_idx(self):
        files = [
            opj(self.root, f"{self.file_basename}_{pdb_id}.h5")
            for pdb_id in self.processed.keys()
        ]
        self.idx = []
        for i, (pdb, group_info) in enumerate(self.processed.items()):
            for temp, replica, num_frames in group_info:
                # build the catalog here for each conformer
                d = [
                    (pdb, files[i], temp, replica, conf_id)
                    for conf_id in range(num_frames)
                ]
                self.idx.extend(d)

        assert (
            len(self.idx) == self.num_conformers
        ), f"Mismatch between number of conformers and idxs: {self.num_conformers} vs {len(self.idx)}"

    def process_specific_group(self, pdb, file, temp, repl, conf_idx):
        # do not use attributes from h5group because is will cause memory leak
        # use the read_direct and np.s_ to get the coords and forces of interest directly
        conf_idx = conf_idx * self.skip_frames
        slice_idxs = np.s_[conf_idx : conf_idx + 1]
        with h5py.File(file, "r") as f:
            z = f[pdb]["z"][:]
            coords = np.zeros((z.shape[0], 3))
            forces = np.zeros((z.shape[0], 3))

            group = f[f"{pdb}/{temp}/{repl}"]

            group["coords"].read_direct(coords, slice_idxs)
            group["forces"].read_direct(forces, slice_idxs)

            # coords and forces shape (num_atoms, 3)
            assert (
                coords.shape[0] == forces.shape[0]
            ), f"Number of frames mismatch between coords and forces: {group['coords'].shape[0]} vs {group['forces'].shape[0]}"
            assert (
                coords.shape[0] == z.shape[0]
            ), f"Number of atoms mismatch between coords and z: {group['coords'].shape[1]} vs {z.shape[0]}"

        return (z, coords, forces)

    def get(self, element):
        data = Data()
        if self.idx is None:
            # this process will be performed, num_workers * num_gpus (one per thread)
            self._setup_idx()
        # fields_data is a tuple with the file, pdb, temp, replica, conf_idx
        pdb_id, file_path, temp, replica, conf_idx = self.idx[element]
        z, coords, forces = self.process_specific_group(
            pdb_id, file_path, temp, replica, conf_idx
        )
        data.z = torch.tensor(z, dtype=torch.long)
        data.pos = torch.tensor(coords, dtype=torch.float)
        data.neg_dy = torch.tensor(forces, dtype=torch.float)
        data.info = f"{pdb_id}_{temp}_{replica}_{conf_idx}"
        return data



# Copyright Universitat Pompeu Fabra 2020-2023  https://www.compscience.org
# Distributed under the MIT License.
# (See accompanying file README.md file or copy at http://opensource.org/licenses/MIT)

from .ace import Ace, AceHF
from .ani import ANI1, ANI1CCX, ANI1X, ANI2X
from .comp6 import (
    ANIMD,
    DrugBank,
    GDB07to09,
    GDB10to13,
    Tripeptides,
    S66X8,
    COMP6v1,
    COMP6v2,
)
from .mdcath import MDCATH
from .custom import Custom
from .water import WaterBox
from .hdf import HDF5
from .md17 import MD17
from .md22 import MD22
from .qm9 import QM9
from .qm9q import QM9q
from .spice import SPICE
from .genentech import GenentechTorsions
from .maceoff import MACEOFF

__all__ = [
    "Ace",
    "AceHF",
    "ANIMD",
    "ANI1",
    "ANI1CCX",
    "ANI1X",
    "ANI2X",
    "COMP6v1",
    "COMP6v2",
    "Custom",
    "DrugBank",
    "GDB07to09",
    "GDB10to13",
    "GenentechTorsions",
    "HDF5",
    "MDCATH",
    "MD17",
    "MD22",
    "QM9",
    "QM9q",
    "S66X8",
    "SPICE",
    "Tripeptides",
    "WaterBox",
    "MACEOFF",
]



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""Warp kernels for computing 3x3 tensor norms (I, A, S components)."""

import warp as wp

from .utils import add_module, get_wp_fp_dtype


def generate_tensor_norm3(dtype: str, h_last: bool = True, use_irmem: bool = True):
    """Generate Warp kernels for computing squared norms of 3x3 tensor I, A, S components."""
    dtype_wp = get_wp_fp_dtype(dtype)

    if not use_irmem:
        raise ValueError(f"only supporting use_irmem True, but got {use_irmem}")

    if not h_last:
        raise ValueError(f"only supporting h_last True but got {h_last}")

    class mat3x3(wp.types.matrix(shape=(3, 3), dtype=dtype_wp)):
        pass

    def tensor_norm3_fwd(
        X: wp.array(ndim=4, dtype=dtype_wp),
        output: wp.array(ndim=2, dtype=dtype_wp),
    ):
        """Computes I, A, S norms of 3x3 tensor: trace², antisym², sym_traceless²."""
        b, h = wp.tid()

        x00 = X[b, 0, 0, h]
        x01 = X[b, 0, 1, h]
        x02 = X[b, 0, 2, h]
        x10 = X[b, 1, 0, h]
        x11 = X[b, 1, 1, h]
        x12 = X[b, 1, 2, h]
        x20 = X[b, 2, 0, h]
        x21 = X[b, 2, 1, h]
        x22 = X[b, 2, 2, h]

        one_half = X.dtype(0.5)
        one_third = X.dtype(1.0 / 3.0)

        trace = x00 + x11 + x22
        trace_third = trace / X.dtype(3.0)
        norm2_i = one_third * trace * trace
        norm2_a = one_half * ((x01 - x10) * (x01 - x10) + (x02 - x20) * (x02 - x20) + (x12 - x21) * (x12 - x21))
        norm2_s = (
            one_half * ((x01 + x10) * (x01 + x10) + (x02 + x20) * (x02 + x20) + (x12 + x21) * (x12 + x21))
            + (x00 - trace_third) * (x00 - trace_third)
            + (x11 - trace_third) * (x11 - trace_third)
            + (x22 - trace_third) * (x22 - trace_third)
        )

        output[b, h] = norm2_i
        output[b, h + X.shape[3]] = norm2_a
        output[b, h + 2 * X.shape[3]] = norm2_s

    def tensor_norm3_bwd(
        grad_output: wp.array(ndim=2, dtype=dtype_wp),
        X: wp.array(ndim=4, dtype=dtype_wp),
        grad_X: wp.array(ndim=4, dtype=dtype_wp),
    ):
        """Backward: grad_X = d(I,A,S norms)/dX · grad_output."""
        b, h = wp.tid()

        grad_i = grad_output[b, h]
        grad_a = grad_output[b, h + X.shape[3]]
        grad_s = grad_output[b, h + 2 * X.shape[3]]

        x00 = X[b, 0, 0, h]
        x01 = X[b, 0, 1, h]
        x02 = X[b, 0, 2, h]
        x10 = X[b, 1, 0, h]
        x11 = X[b, 1, 1, h]
        x12 = X[b, 1, 2, h]
        x20 = X[b, 2, 0, h]
        x21 = X[b, 2, 1, h]
        x22 = X[b, 2, 2, h]

        trace = x00 + x11 + x22
        trace_third = trace / X.dtype(3.0)

        diag_grad_i = X.dtype(2.0 / 3.0) * trace * grad_i

        dev00 = x00 - trace_third
        dev11 = x11 - trace_third
        dev22 = x22 - trace_third

        c4_3 = X.dtype(4.0) / X.dtype(3.0)
        c2_3 = X.dtype(2.0) / X.dtype(3.0)

        grad_s_term_00 = c4_3 * dev00 - c2_3 * dev11 - c2_3 * dev22
        grad_s_term_11 = c4_3 * dev11 - c2_3 * dev00 - c2_3 * dev22
        grad_s_term_22 = c4_3 * dev22 - c2_3 * dev00 - c2_3 * dev11

        grad_X[b, 0, 0, h] = diag_grad_i + grad_s * grad_s_term_00
        grad_X[b, 1, 1, h] = diag_grad_i + grad_s * grad_s_term_11
        grad_X[b, 2, 2, h] = diag_grad_i + grad_s * grad_s_term_22

        diff01 = x01 - x10
        sum01 = x01 + x10
        grad_X[b, 0, 1, h] = grad_a * diff01 + grad_s * sum01
        grad_X[b, 1, 0, h] = -grad_a * diff01 + grad_s * sum01

        diff02 = x02 - x20
        sum02 = x02 + x20
        grad_X[b, 0, 2, h] = grad_a * diff02 + grad_s * sum02
        grad_X[b, 2, 0, h] = -grad_a * diff02 + grad_s * sum02

        diff12 = x12 - x21
        sum12 = x12 + x21
        grad_X[b, 1, 2, h] = grad_a * diff12 + grad_s * sum12
        grad_X[b, 2, 1, h] = -grad_a * diff12 + grad_s * sum12

    def tensor_norm3_bwd_bwd(
        grad_grad_X: wp.array(ndim=4, dtype=dtype_wp),
        X: wp.array(ndim=4, dtype=dtype_wp),
        grad_output: wp.array(ndim=2, dtype=dtype_wp),
        grad_grad_output: wp.array(ndim=2, dtype=dtype_wp),
        grad_x: wp.array(ndim=4, dtype=dtype_wp),
    ):
        """Computes d(grad_X)/d(grad_output) and d(grad_X)/d(X) contracted with grad_grad_X."""
        b, h = wp.tid()

        gg00 = grad_grad_X[b, 0, 0, h]
        gg01 = grad_grad_X[b, 0, 1, h]
        gg02 = grad_grad_X[b, 0, 2, h]
        gg10 = grad_grad_X[b, 1, 0, h]
        gg11 = grad_grad_X[b, 1, 1, h]
        gg12 = grad_grad_X[b, 1, 2, h]
        gg20 = grad_grad_X[b, 2, 0, h]
        gg21 = grad_grad_X[b, 2, 1, h]
        gg22 = grad_grad_X[b, 2, 2, h]

        x00 = X[b, 0, 0, h]
        x01 = X[b, 0, 1, h]
        x02 = X[b, 0, 2, h]
        x10 = X[b, 1, 0, h]
        x11 = X[b, 1, 1, h]
        x12 = X[b, 1, 2, h]
        x20 = X[b, 2, 0, h]
        x21 = X[b, 2, 1, h]
        x22 = X[b, 2, 2, h]

        grad_i = grad_output[b, h]
        grad_a = grad_output[b, h + X.shape[3]]
        grad_s = grad_output[b, h + 2 * X.shape[3]]

        trace_X = x00 + x11 + x22
        trace_gg = gg00 + gg11 + gg22
        c2_3 = X.dtype(2.0 / 3.0)
        c4_3 = X.dtype(4.0 / 3.0)

        # Part 1: grad_grad_output = d(grad_X)/d(grad_output) · grad_grad_X
        # I channel: (2/3) * trace(X) * trace(gg)
        grad_grad_output[b, h] = c2_3 * trace_X * trace_gg

        # A channel: diff_X · diff_gg
        diff01_X = x01 - x10
        diff02_X = x02 - x20
        diff12_X = x12 - x21
        diff01_gg = gg01 - gg10
        diff02_gg = gg02 - gg20
        diff12_gg = gg12 - gg21
        grad_grad_output[b, h + X.shape[3]] = diff01_X * diff01_gg + diff02_X * diff02_gg + diff12_X * diff12_gg

        # S channel: sum_X · sum_gg + dev_terms · diag_gg
        trace_third_X = trace_X / X.dtype(3.0)
        dev00 = x00 - trace_third_X
        dev11 = x11 - trace_third_X
        dev22 = x22 - trace_third_X
        grad_s_term_00 = c4_3 * dev00 - c2_3 * dev11 - c2_3 * dev22
        grad_s_term_11 = c4_3 * dev11 - c2_3 * dev00 - c2_3 * dev22
        grad_s_term_22 = c4_3 * dev22 - c2_3 * dev00 - c2_3 * dev11
        sum01_X = x01 + x10
        sum02_X = x02 + x20
        sum12_X = x12 + x21
        sum01_gg = gg01 + gg10
        sum02_gg = gg02 + gg20
        sum12_gg = gg12 + gg21
        grad_grad_output_s = sum01_X * sum01_gg + sum02_X * sum02_gg + sum12_X * sum12_gg
        grad_grad_output_s += grad_s_term_00 * gg00 + grad_s_term_11 * gg11 + grad_s_term_22 * gg22
        grad_grad_output[b, h + 2 * X.shape[3]] = grad_grad_output_s

        # Part 2: grad_x = d(grad_X)/d(X) · grad_grad_X
        # I channel: (2/3) * grad_i * trace(gg) on diagonals
        scalar_diag = c2_3 * grad_i * trace_gg

        # A channel: grad_a * diff_gg (antisymmetric)
        antisym_01 = grad_a * diff01_gg
        antisym_02 = grad_a * diff02_gg
        antisym_12 = grad_a * diff12_gg

        # S channel off-diag: grad_s * sum_gg
        sym_offdiag_01 = grad_s * sum01_gg
        sym_offdiag_02 = grad_s * sum02_gg
        sym_offdiag_12 = grad_s * sum12_gg

        # S channel diag: grad_s * (4/3 on self, -2/3 on others)
        sym_diag_00 = grad_s * (c4_3 * gg00 - c2_3 * gg11 - c2_3 * gg22)
        sym_diag_11 = grad_s * (c4_3 * gg11 - c2_3 * gg00 - c2_3 * gg22)
        sym_diag_22 = grad_s * (c4_3 * gg22 - c2_3 * gg00 - c2_3 * gg11)

        # Diagonals
        grad_x[b, 0, 0, h] = scalar_diag + sym_diag_00
        grad_x[b, 1, 1, h] = scalar_diag + sym_diag_11
        grad_x[b, 2, 2, h] = scalar_diag + sym_diag_22

        # Off-diagonals
        grad_x[b, 0, 1, h] = antisym_01 + sym_offdiag_01
        grad_x[b, 1, 0, h] = -antisym_01 + sym_offdiag_01
        grad_x[b, 0, 2, h] = antisym_02 + sym_offdiag_02
        grad_x[b, 2, 0, h] = -antisym_02 + sym_offdiag_02
        grad_x[b, 1, 2, h] = antisym_12 + sym_offdiag_12
        grad_x[b, 2, 1, h] = -antisym_12 + sym_offdiag_12

    return (
        wp.Kernel(
            tensor_norm3_fwd,
            key=f"tensor_norm3_fwd_{dtype}",
            module=wp.get_module(f"tensor_norm3_fwd_{dtype}"),
        ),
        wp.Kernel(
            tensor_norm3_bwd,
            key=f"tensor_norm3_bwd_{dtype}",
            module=wp.get_module(f"tensor_norm3_bwd_{dtype}"),
        ),
        wp.Kernel(
            tensor_norm3_bwd_bwd,
            key=f"tensor_norm3_bwd_bwd_{dtype}",
            module=wp.get_module(f"tensor_norm3_bwd_bwd_{dtype}"),
        ),
    )


tensor_norm3_fwd_fp64, tensor_norm3_bwd_fp64, tensor_norm3_bwd_bwd_fp64 = generate_tensor_norm3("float64")
tensor_norm3_fwd_fp32, tensor_norm3_bwd_fp32, tensor_norm3_bwd_bwd_fp32 = generate_tensor_norm3("float32")
tensor_norm3_fwd_fp16, tensor_norm3_bwd_fp16, tensor_norm3_bwd_bwd_fp16 = generate_tensor_norm3("float16")

add_module("tensor_norm3_fwd", ["float64"], tensor_norm3_fwd_fp64)
add_module("tensor_norm3_bwd", ["float64"], tensor_norm3_bwd_fp64)
add_module("tensor_norm3_bwd_bwd", ["float64"], tensor_norm3_bwd_bwd_fp64)

add_module("tensor_norm3_fwd", ["float32"], tensor_norm3_fwd_fp32)
add_module("tensor_norm3_bwd", ["float32"], tensor_norm3_bwd_fp32)
add_module("tensor_norm3_bwd_bwd", ["float32"], tensor_norm3_bwd_bwd_fp32)

add_module("tensor_norm3_fwd", ["float16"], tensor_norm3_fwd_fp16)
add_module("tensor_norm3_bwd", ["float16"], tensor_norm3_bwd_fp16)
add_module("tensor_norm3_bwd_bwd", ["float16"], tensor_norm3_bwd_bwd_fp16)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""Custom tensor operations with Warp kernel implementations."""

from __future__ import annotations

import warp as wp

from .compose_tensor import fn_compose_tensor
from .decompose_tensor import fn_decompose_tensor
from .equivariant_o3_matmul import fn_tensor_matmul_o3_3x3
from .equivariant_so3_matmul import fn_tensor_matmul_so3_3x3
from .graph_transform import graph_transform
from .tensor_norm3 import fn_tensor_norm3
from .tensornet_mp import fn_message_passing
from .tensornet_radial_mp import fn_radial_message_passing

wp.init()

__all__ = [
    "fn_compose_tensor",
    "fn_decompose_tensor",
    "fn_message_passing",
    "fn_radial_message_passing",
    "fn_radial_message_passing",
    "fn_tensor_matmul_o3_3x3",
    "fn_tensor_matmul_so3_3x3",
    "fn_tensor_norm3",
    "graph_transform",
]



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from __future__ import annotations

import torch
import warp as wp
from torch import Tensor

from torchmdnet.extensions.warp_kernels import (
    convert_to_sparse,
    count_row_col,
    get_stream,
)


@torch.library.custom_op(
    "nvtnet::count_row_col_primitive",
    mutates_args=(),
    device_types=["cpu", "cuda"],
)
def _(edge_index: Tensor, num_nodes: int) -> tuple[Tensor, Tensor]:
    stream = get_stream(edge_index.device)
    device = wp.device_from_torch(edge_index.device)
    row_count = torch.zeros(num_nodes + 1, dtype=torch.int32, device=edge_index.device)
    col_count = torch.zeros(num_nodes + 1, dtype=torch.int32, device=edge_index.device)

    edge_index_wp = wp.from_torch(edge_index, return_ctype=True)
    row_count_wp = wp.from_torch(row_count, return_ctype=True)
    col_count_wp = wp.from_torch(col_count, return_ctype=True)

    wp.launch(
        count_row_col,
        dim=(edge_index.shape[1]),
        stream=stream,
        device=device,
        inputs=(edge_index_wp, row_count_wp, col_count_wp),
    )

    return row_count, col_count


@torch.library.register_fake("nvtnet::count_row_col_primitive")
def _(edge_index: Tensor, num_nodes: int) -> tuple[Tensor, Tensor]:
    output = torch.zeros(num_nodes + 1, dtype=torch.int32, device=edge_index.device)
    output2 = torch.zeros(num_nodes + 1, dtype=torch.int32, device=edge_index.device)
    return output, output2


@torch.library.custom_op(
    "nvtnet::convert_to_sparse_primitive",
    mutates_args=(),
    device_types=["cpu", "cuda"],
)
def _(
    edge_index: Tensor,
    row_count: Tensor,
    col_count: Tensor,
    row_indptr: Tensor,
    col_indptr: Tensor,
) -> tuple[Tensor, Tensor, Tensor, Tensor]:
    stream = get_stream(edge_index.device)
    device = wp.device_from_torch(edge_index.device)
    edge_index_wp = wp.from_torch(edge_index, return_ctype=True)

    row_count_wp = wp.from_torch(row_count, return_ctype=True)
    col_count_wp = wp.from_torch(col_count, return_ctype=True)

    row_indptr_wp = wp.from_torch(row_indptr, return_ctype=True)
    col_indptr_wp = wp.from_torch(col_indptr, return_ctype=True)

    row_indices = torch.empty(
        edge_index.shape[1], dtype=torch.int32, device=edge_index.device
    )
    col_indices = torch.empty(
        edge_index.shape[1], dtype=torch.int32, device=edge_index.device
    )

    row_data = torch.empty(
        edge_index.shape[1], dtype=torch.int32, device=edge_index.device
    )
    col_data = torch.empty(
        edge_index.shape[1], dtype=torch.int32, device=edge_index.device
    )

    row_indices_wp = wp.from_torch(row_indices, return_ctype=True)
    col_indices_wp = wp.from_torch(col_indices, return_ctype=True)

    row_data_wp = wp.from_torch(row_data, return_ctype=True)
    col_data_wp = wp.from_torch(col_data, return_ctype=True)

    wp.launch(
        convert_to_sparse,
        dim=(edge_index.shape[1]),
        stream=stream,
        device=device,
        inputs=(
            edge_index_wp,
            row_count_wp,
            col_count_wp,
            row_indptr_wp,
            col_indptr_wp,
            row_indices_wp,
            col_indices_wp,
            row_data_wp,
            col_data_wp,
        ),
    )

    return row_indices, col_indices, row_data, col_data


@torch.library.register_fake("nvtnet::convert_to_sparse_primitive")
def _(
    edge_index: Tensor,
    row_count: Tensor,
    col_count: Tensor,
    row_indptr: Tensor,
    col_indptr: Tensor,
) -> tuple[Tensor, Tensor, Tensor, Tensor]:
    output = torch.empty(
        edge_index.shape[1], dtype=torch.int32, device=edge_index.device
    )
    output2 = torch.empty(
        edge_index.shape[1], dtype=torch.int32, device=edge_index.device
    )
    output3 = torch.empty(
        edge_index.shape[1], dtype=torch.int32, device=edge_index.device
    )
    output4 = torch.empty(
        edge_index.shape[1], dtype=torch.int32, device=edge_index.device
    )
    return output, output2, output3, output4


@torch.compiler.allow_in_graph
def graph_transform(
    edge_index: Tensor, num_nodes: int
) -> tuple[Tensor, Tensor, Tensor, Tensor, Tensor, Tensor]:
    row_count, col_count = torch.ops.nvtnet.count_row_col_primitive(
        edge_index, num_nodes
    )
    row_indptr, col_indptr = (
        torch.cumsum(row_count, dim=0, dtype=torch.int32),
        torch.cumsum(col_count, dim=0, dtype=torch.int32),
    )
    (
        row_indices,
        col_indices,
        row_data,
        col_data,
    ) = torch.ops.nvtnet.convert_to_sparse_primitive(
        edge_index, row_count, col_count, row_indptr, col_indptr
    )
    return row_data, row_indices, row_indptr, col_data, col_indices, col_indptr



# Copyright Universitat Pompeu Fabra 2020-2023  https://www.compscience.org
# Distributed under the MIT License.
# (See accompanying file README.md file or copy at http://opensource.org/licenses/MIT)
import triton
import triton.language as tl
from torch import Tensor
import torch
from torchmdnet.extensions.triton_neighbors import (
    _tl_round,
    TritonNeighborAutograd,
    neighbor_grad_positions,
    neighbor_op_setup_context,
)
from torch.library import triton_op, wrap_triton
from typing import Tuple


@triton.jit
def _neighbor_brute_kernel(
    pos_ptr,
    batch_ptr,
    box_ptr,
    neighbors0_ptr,
    neighbors1_ptr,
    deltas_ptr,
    distances_ptr,
    counter_ptr,
    box_batch_stride,
    n_atoms,
    num_all_pairs,
    use_periodic: tl.constexpr,
    include_transpose: tl.constexpr,
    loop: tl.constexpr,
    max_pairs,
    cutoff_lower,
    cutoff_upper,
    BLOCK: tl.constexpr,
):
    """Brute-force neighbor list kernel using triangular indexing and atomic compaction.

    Uses triangular indexing to iterate over only n*(n-1)/2 pairs (or n*(n+1)/2 with loop),
    achieving 100% thread utilization while maintaining block-level atomic compaction.
    """
    pid = tl.program_id(axis=0)
    start = pid * BLOCK
    idx = start + tl.arange(0, BLOCK)

    valid = idx < num_all_pairs

    # Convert linear index to (i, j) using triangular formula (same as CUDA get_row)
    # Do integer arithmetic first, only convert to float for sqrt
    if loop:
        # With self-loops: j <= i, num_pairs = n*(n+1)/2
        # row = floor((-1 + sqrt(1 + 8k)) / 2)
        sqrt_arg = (1 + 8 * idx).to(tl.float32)
        row_f = tl.math.floor((-1.0 + tl.math.sqrt(sqrt_arg)) * 0.5)
        i = row_f.to(tl.int32)
        # Handle floating-point edge case: if i*(i+1)/2 > idx, decrement i
        i = tl.where(i * (i + 1) > 2 * idx, i - 1, i)
        # col = k - row*(row+1)/2
        j = idx - (i * (i + 1)) // 2
    else:
        # Without self-loops: j < i, num_pairs = n*(n-1)/2
        # row = floor((1 + sqrt(1 + 8k)) / 2)
        sqrt_arg = (1 + 8 * idx).to(tl.float32)
        row_f = tl.math.floor((1.0 + tl.math.sqrt(sqrt_arg)) * 0.5)
        i = row_f.to(tl.int32)
        # Handle floating-point edge case (same correction as CUDA)
        i = tl.where(i * (i - 1) > 2 * idx, i - 1, i)
        # col = k - row*(row-1)/2
        j = idx - (i * (i - 1)) // 2

    # Validate indices: check bounds and triangular constraint
    # Due to float precision, we may get invalid (i, j) pairs
    valid = valid & (i >= 0) & (i < n_atoms) & (j >= 0) & (j <= i)
    if not loop:
        # For non-loop case, also require j < i (no self-loops)
        valid = valid & (j < i)

    batch_i = tl.load(batch_ptr + i, mask=valid, other=0)
    batch_j = tl.load(batch_ptr + j, mask=valid, other=0)
    valid = valid & (batch_i == batch_j)

    pos_ix = tl.load(pos_ptr + i * 3 + 0, mask=valid, other=0.0)
    pos_iy = tl.load(pos_ptr + i * 3 + 1, mask=valid, other=0.0)
    pos_iz = tl.load(pos_ptr + i * 3 + 2, mask=valid, other=0.0)
    pos_jx = tl.load(pos_ptr + j * 3 + 0, mask=valid, other=0.0)
    pos_jy = tl.load(pos_ptr + j * 3 + 1, mask=valid, other=0.0)
    pos_jz = tl.load(pos_ptr + j * 3 + 2, mask=valid, other=0.0)

    dx = pos_ix - pos_jx
    dy = pos_iy - pos_jy
    dz = pos_iz - pos_jz

    if use_periodic:
        box_base = box_ptr + batch_i * box_batch_stride

        b20 = tl.load(box_base + 2 * 3 + 0, mask=valid, other=0.0)
        b21 = tl.load(box_base + 2 * 3 + 1, mask=valid, other=0.0)
        b22 = tl.load(box_base + 2 * 3 + 2, mask=valid, other=1.0)
        b10 = tl.load(box_base + 1 * 3 + 0, mask=valid, other=0.0)
        b11 = tl.load(box_base + 1 * 3 + 1, mask=valid, other=1.0)
        b00 = tl.load(box_base + 0 * 3 + 0, mask=valid, other=1.0)

        scale3 = _tl_round(dz / b22)
        dx = dx - scale3 * b20
        dy = dy - scale3 * b21
        dz = dz - scale3 * b22
        scale2 = _tl_round(dy / b11)
        dx = dx - scale2 * b10
        dy = dy - scale2 * b11
        scale1 = _tl_round(dx / b00)
        dx = dx - scale1 * b00

    dist2 = dx * dx + dy * dy + dz * dz
    dist = tl.sqrt(dist2)
    # Self-loops (i == j) are exempt from cutoff_lower since they have distance 0
    is_self_loop = i == j
    valid = valid & (dist < cutoff_upper) & ((dist >= cutoff_lower) | is_self_loop)

    valid_int = valid.to(tl.int32)
    local_idx = tl.cumsum(valid_int, axis=0) - 1
    total_pairs = tl.sum(valid_int, axis=0)

    # For transpose, don't count self-loops (i == j)
    if include_transpose:
        is_not_self_loop = i != j
        valid_transpose = valid & is_not_self_loop
        total_transpose = tl.sum(valid_transpose.to(tl.int32), axis=0)
        total_out = total_pairs + total_transpose
    else:
        total_out = total_pairs

    has_work = total_out > 0
    start_idx = tl.atomic_add(counter_ptr, total_out, mask=has_work)
    start_idx = tl.where(has_work, start_idx, 0)
    start_idx = tl.broadcast_to(start_idx, local_idx.shape)

    write_idx = start_idx + local_idx
    mask_store = valid & (write_idx < max_pairs)
    tl.store(neighbors0_ptr + write_idx, i, mask=mask_store)
    tl.store(neighbors1_ptr + write_idx, j, mask=mask_store)
    tl.store(deltas_ptr + write_idx * 3 + 0, dx, mask=mask_store)
    tl.store(deltas_ptr + write_idx * 3 + 1, dy, mask=mask_store)
    tl.store(deltas_ptr + write_idx * 3 + 2, dz, mask=mask_store)
    tl.store(distances_ptr + write_idx, dist, mask=mask_store)

    if include_transpose:
        # Don't add transpose for self-loops (i == j)
        is_not_self_loop = i != j
        valid_transpose = valid & is_not_self_loop
        valid_t_int = valid_transpose.to(tl.int32)
        local_idx_t = tl.cumsum(valid_t_int, axis=0) - 1
        total_pairs_t = tl.sum(valid_t_int, axis=0)

        write_idx_t = start_idx + total_pairs + local_idx_t
        mask_store_t = valid_transpose & (write_idx_t < max_pairs)
        tl.store(neighbors0_ptr + write_idx_t, j, mask=mask_store_t)
        tl.store(neighbors1_ptr + write_idx_t, i, mask=mask_store_t)
        tl.store(deltas_ptr + write_idx_t * 3 + 0, -dx, mask=mask_store_t)
        tl.store(deltas_ptr + write_idx_t * 3 + 1, -dy, mask=mask_store_t)
        tl.store(deltas_ptr + write_idx_t * 3 + 2, -dz, mask=mask_store_t)
        tl.store(distances_ptr + write_idx_t, dist, mask=mask_store_t)


@triton_op("torchmdnet::triton_neighbor_bruteforce", mutates_args={})
def triton_neighbor_bruteforce(
    positions: Tensor,
    batch: Tensor,
    box_vectors: Tensor,
    use_periodic: bool,
    cutoff_lower: float,
    cutoff_upper: float,
    max_num_pairs: int,
    loop: bool,
    include_transpose: bool,
) -> Tuple[Tensor, Tensor, Tensor, Tensor]:
    device = positions.device
    dtype = positions.dtype
    n_atoms = positions.size(0)

    batch = batch.contiguous()
    positions = positions.contiguous()
    if use_periodic:
        if box_vectors.dim() == 2:
            box_vectors = box_vectors.unsqueeze(0)
        elif box_vectors.dim() != 3:
            raise ValueError('Expected "box_vectors" to have shape (n_batch, 3, 3)')
        box_vectors = box_vectors.to(device=device, dtype=dtype)
        box_vectors = box_vectors.contiguous()
        # Use stride 0 to broadcast single box to all batches (avoids CPU sync)
        box_batch_stride = 0 if box_vectors.size(0) == 1 else 9

    neighbors = torch.full((2, max_num_pairs), -1, device=device, dtype=torch.long)
    deltas = torch.zeros((max_num_pairs, 3), device=device, dtype=dtype)
    distances = torch.zeros((max_num_pairs,), device=device, dtype=dtype)
    num_pairs = torch.zeros((1,), device=device, dtype=torch.int32)

    # Compute triangular pair count: n*(n-1)/2 without self-loops, n*(n+1)/2 with
    if loop:
        num_all_pairs = n_atoms * (n_atoms + 1) // 2
    else:
        num_all_pairs = n_atoms * (n_atoms - 1) // 2

    # Grid covers only triangular pairs (not n*n)
    grid = lambda meta: (triton.cdiv(num_all_pairs, meta["BLOCK"]),)

    wrap_triton(_neighbor_brute_kernel)[grid](
        positions,
        batch,
        box_vectors if use_periodic else positions,  # dummy pointer if not periodic
        neighbors[0],
        neighbors[1],
        deltas,
        distances,
        num_pairs,
        box_batch_stride if use_periodic else 0,
        n_atoms,
        num_all_pairs,
        use_periodic,
        include_transpose,
        loop,
        max_num_pairs,
        cutoff_lower,
        cutoff_upper,
        BLOCK=256,
    )

    return neighbors, deltas, distances, num_pairs


def _bruteforce_backward(
    ctx, grad_neighbors, grad_deltas, grad_distances, grad_num_pairs
):
    grad_positions = neighbor_grad_positions(ctx, grad_deltas, grad_distances)
    return grad_positions, None, None, None, None, None, None, None, None


torch.library.register_autograd(
    "torchmdnet::triton_neighbor_bruteforce",
    _bruteforce_backward,
    setup_context=neighbor_op_setup_context,
)


class TritonBruteNeighborAutograd(TritonNeighborAutograd):
    @staticmethod
    def forward(  # type: ignore[override]
        ctx,
        positions: Tensor,
        batch: Tensor,
        box_vectors: Tensor,
        use_periodic: bool,
        cutoff_lower: float,
        cutoff_upper: float,
        max_num_pairs: int,
        loop: bool,
        include_transpose: bool,
    ):
        neighbors, deltas, distances, num_pairs = triton_neighbor_bruteforce(
            positions,
            batch,
            box_vectors,
            use_periodic,
            cutoff_lower,
            cutoff_upper,
            max_num_pairs,
            loop,
            include_transpose,
        )

        ctx.save_for_backward(neighbors, deltas, distances)
        ctx.num_atoms = positions.size(0)
        return neighbors, deltas, distances, num_pairs



# Copyright Universitat Pompeu Fabra 2020-2023  https://www.compscience.org
# Distributed under the MIT License.
# (See accompanying file README.md file or copy at http://opensource.org/licenses/MIT)

import torch
from torch import Tensor
from typing import Tuple


def torch_neighbor_bruteforce_memory_efficient(
    strategy: str,
    positions: Tensor,
    batch: Tensor,
    in_box_vectors: Tensor,
    use_periodic: bool,
    cutoff_lower: float,
    cutoff_upper: float,
    max_num_pairs: int,
    loop: bool,
    include_transpose: bool,
) -> Tuple[Tensor, Tensor, Tensor, Tensor]:
    """
    Compute neighbor pairs within a cutoff distance using a more memory-efficient approach but not torch.compile compatible.

    Args:
        strategy: Strategy string (unused in CPU implementation)
        positions: Tensor of shape (n_atoms, 3) with atom positions
        batch: Tensor of shape (n_atoms,) with batch indices for each atom
        in_box_vectors: Box vectors for periodic boundary conditions, shape (3, 3) or (n_batch, 3, 3)
        use_periodic: Whether to use periodic boundary conditions
        cutoff_lower: Lower cutoff distance
        cutoff_upper: Upper cutoff distance
        max_num_pairs: Maximum number of pairs to return
        loop: Whether to include self-interactions (i, i)
        include_transpose: Whether to include both (i, j) and (j, i) pairs

    Returns:
        neighbors: Tensor of shape (2, num_pairs) with neighbor indices
        deltas: Tensor of shape (num_pairs, 3) with displacement vectors
        distances: Tensor of shape (num_pairs,) with distances
        num_pairs_found: Tensor of shape (1,) with the number of pairs found
    """
    from torchmdnet.extensions.neighbors import _round_nearest

    assert positions.dim() == 2, 'Expected "positions" to have two dimensions'
    assert (
        positions.size(0) > 0
    ), 'Expected the 1st dimension size of "positions" to be more than 0'
    assert (
        positions.size(1) == 3
    ), 'Expected the 2nd dimension size of "positions" to be 3'
    assert positions.is_contiguous(), 'Expected "positions" to be contiguous'
    assert cutoff_upper > 0, 'Expected "cutoff" to be positive'

    box_vectors = in_box_vectors
    n_batch = batch.max().item() + 1

    if use_periodic:
        if box_vectors.dim() == 2:
            box_vectors = box_vectors.unsqueeze(0).expand(n_batch, 3, 3)

        assert box_vectors.dim() == 3, 'Expected "box_vectors" to have three dimensions'
        assert (
            box_vectors.size(1) == 3 and box_vectors.size(2) == 3
        ), 'Expected "box_vectors" to have shape (n_batch, 3, 3)'
        assert (
            box_vectors.size(0) == n_batch
        ), 'Expected "box_vectors" to have shape (n_batch, 3, 3)'

        # Check that the box is a valid triclinic box, only check the first one
        v = box_vectors[0]
        c = cutoff_upper

        assert v[0, 1] == 0, "Invalid box vectors: box_vectors[0][1] != 0"
        assert v[0, 2] == 0, "Invalid box vectors: box_vectors[0][2] != 0"
        assert v[1, 2] == 0, "Invalid box vectors: box_vectors[1][2] != 0"
        assert v[0, 0] >= 2 * c, "Invalid box vectors: box_vectors[0][0] < 2*cutoff"
        assert v[1, 1] >= 2 * c, "Invalid box vectors: box_vectors[1][1] < 2*cutoff"
        assert v[2, 2] >= 2 * c, "Invalid box vectors: box_vectors[2][2] < 2*cutoff"
        assert (
            v[0, 0] >= 2 * v[1, 0]
        ), "Invalid box vectors: box_vectors[0][0] < 2*box_vectors[1][0]"
        assert (
            v[0, 0] >= 2 * v[2, 0]
        ), "Invalid box vectors: box_vectors[0][0] < 2*box_vectors[2][0]"
        assert (
            v[1, 1] >= 2 * v[2, 1]
        ), "Invalid box vectors: box_vectors[1][1] < 2*box_vectors[2][1]"

    assert max_num_pairs > 0, 'Expected "max_num_neighbors" to be positive'

    n_atoms = positions.size(0)

    # Generate all pairs using lower triangular indices
    neighbors = torch.tril_indices(
        n_atoms, n_atoms, offset=-1, device=positions.device, dtype=torch.int32
    )

    # Filter pairs to same batch
    mask = batch[neighbors[0]] == batch[neighbors[1]]
    neighbors = neighbors[:, mask].to(torch.int32)

    # Calculate deltas
    deltas = positions[neighbors[0]] - positions[neighbors[1]]

    if use_periodic:
        pair_batch = batch[neighbors[0]]

        # Apply periodic boundary conditions for triclinic box
        scale3 = _round_nearest(deltas[:, 2] / box_vectors[pair_batch, 2, 2])
        deltas[:, 0] = deltas[:, 0] - scale3 * box_vectors[pair_batch, 2, 0]
        deltas[:, 1] = deltas[:, 1] - scale3 * box_vectors[pair_batch, 2, 1]
        deltas[:, 2] = deltas[:, 2] - scale3 * box_vectors[pair_batch, 2, 2]

        scale2 = _round_nearest(deltas[:, 1] / box_vectors[pair_batch, 1, 1])
        deltas[:, 0] = deltas[:, 0] - scale2 * box_vectors[pair_batch, 1, 0]
        deltas[:, 1] = deltas[:, 1] - scale2 * box_vectors[pair_batch, 1, 1]

        scale1 = _round_nearest(deltas[:, 0] / box_vectors[pair_batch, 0, 0])
        deltas[:, 0] = deltas[:, 0] - scale1 * box_vectors[pair_batch, 0, 0]

    # Calculate distances (gradient-safe for zero distances)
    dist_sq = (deltas * deltas).sum(dim=-1)
    zero_mask = dist_sq == 0
    distances = torch.where(
        zero_mask,
        torch.zeros_like(dist_sq),
        torch.sqrt(dist_sq.clamp(min=1e-32)),
    )

    # Filter by cutoff
    mask = (distances < cutoff_upper) & (distances >= cutoff_lower)
    neighbors = neighbors[:, mask]
    deltas = deltas[mask]
    distances = distances[mask]

    if include_transpose:
        neighbors = torch.hstack([neighbors, torch.stack([neighbors[1], neighbors[0]])])
        distances = torch.hstack([distances, distances])
        deltas = torch.vstack([deltas, -deltas])

    if loop:
        range_tensor = torch.arange(
            0, n_atoms, dtype=torch.int32, device=positions.device
        )
        neighbors = torch.hstack([neighbors, torch.stack([range_tensor, range_tensor])])
        distances = torch.hstack(
            [distances, torch.zeros_like(range_tensor, dtype=distances.dtype)]
        )
        deltas = torch.vstack(
            [deltas, torch.zeros(n_atoms, 3, dtype=deltas.dtype, device=deltas.device)]
        )

    num_pairs_found = torch.tensor(
        [distances.size(0)], dtype=torch.int32, device=positions.device
    )

    # Pad to max_num_pairs to enable torch.compile by guaranteeing predictable output size
    extension = max(max_num_pairs - distances.size(0), 0)
    if extension > 0:
        deltas = torch.vstack(
            [
                deltas,
                torch.zeros(extension, 3, dtype=deltas.dtype, device=deltas.device),
            ]
        )
        distances = torch.hstack(
            [
                distances,
                torch.zeros(extension, dtype=distances.dtype, device=distances.device),
            ]
        )
        # For the neighbors add (-1, -1) pairs to fill the tensor
        neighbors = torch.hstack(
            [
                neighbors,
                torch.full(
                    (2, extension), -1, dtype=torch.int32, device=neighbors.device
                ),
            ]
        )

    return neighbors, deltas, distances, num_pairs_found



# Copyright Universitat Pompeu Fabra 2020-2023  https://www.compscience.org
# Distributed under the MIT License.
# (See accompanying file README.md file or copy at http://opensource.org/licenses/MIT)

import h5py
import numpy as np
import torch as pt
from torch_geometric.data import Data, Dataset, download_url, extract_tar
from torchmdnet.datasets.memdataset import MemmappedDataset
from tqdm import tqdm
from torchmdnet.datasets.ani import ANIBase
import os

"""
COmprehensive Machine-learning Potential (COMP6) Benchmark Suite

For more details check:
  - https://arxiv.org/pdf/1801.09319.pdf
  - https://github.com/isayev/COMP6
  - https://aip.scitation.org/doi/abs/10.1063/1.5023802

"""


class COMP6Base(MemmappedDataset):
    _ELEMENT_ENERGIES = {
        1: -0.500607632585,
        6: -37.8302333826,
        7: -54.5680045287,
        8: -75.0362229210,
    }  # Copied from ANI-1x
    ATOMIC_NUMBERS = {b"H": 1, b"C": 6, b"N": 7, b"O": 8}
    HARTREE_TO_EV = 27.211386246

    def __init__(
        self,
        root,
        transform=None,
        pre_transform=None,
        pre_filter=None,
    ):
        self.name = self.__class__.__name__
        super().__init__(
            root,
            transform,
            pre_transform,
            pre_filter,
            properties=("y", "neg_dy"),
        )

    @property
    def raw_url_name(self):
        return self.__class__.__name__

    @property
    def raw_url(self):
        url_prefix = "https://raw.githubusercontent.com/isayev/COMP6/master/COMP6v1"
        return [
            f"{url_prefix}/{self.raw_url_name}/{name}" for name in self.raw_file_names
        ]

    def get_atomref(self, max_z=100):
        """Atomic energy reference values for the :py:mod:`torchmdnet.priors.Atomref` prior.

        Args:
            max_z (int): Maximum atomic number

        Returns:
            torch.Tensor: Atomic energy reference values for each element in the dataset.
        """
        refs = pt.zeros(max_z)
        for key, val in self._ELEMENT_ENERGIES.items():
            refs[key] = val * self.HARTREE_TO_EV

        return refs.view(-1, 1)

    def download(self):
        for url in self.raw_url:
            download_url(url, self.raw_dir)

    def sample_iter(self, mol_ids=False):
        for path in tqdm(self.raw_paths, desc="Files"):
            molecules = list(h5py.File(path).values())[0].items()

            for mol_id, mol in tqdm(molecules, desc="Molecules", leave=False):
                z = pt.tensor(
                    [self.ATOMIC_NUMBERS[atom] for atom in mol["species"]],
                    dtype=pt.long,
                )
                all_pos = pt.tensor(mol["coordinates"][:], dtype=pt.float32)
                all_y = pt.tensor(
                    mol["energies"][:] * self.HARTREE_TO_EV, dtype=pt.float64
                )
                all_neg_dy = pt.tensor(
                    mol["forces"][:] * self.HARTREE_TO_EV, dtype=pt.float32
                )
                all_neg_dy = (
                    -all_neg_dy
                )  # The COMP6 datasets accidentally have gradients as forces

                assert all_pos.shape[0] == all_y.shape[0]
                assert all_pos.shape[1] == z.shape[0]
                assert all_pos.shape[2] == 3

                assert all_neg_dy.shape[0] == all_y.shape[0]
                assert all_neg_dy.shape[1] == z.shape[0]
                assert all_neg_dy.shape[2] == 3

                for pos, y, neg_dy in zip(all_pos, all_y, all_neg_dy):
                    # Create a sample
                    args = dict(z=z, pos=pos, y=y.view(1, 1), neg_dy=neg_dy)
                    if mol_ids:
                        args["mol_id"] = f"{os.path.basename(path)}_{mol_id}"
                    data = Data(**args)

                    if self.pre_filter is not None and not self.pre_filter(data):
                        continue

                    if self.pre_transform is not None:
                        data = self.pre_transform(data)

                    yield data


class ANIMD(COMP6Base):
    """
    ANI Molecular Dynamics (ANI-MD) Benchmark. Forces from the ANI-1x potential are
    applied to run 1ns of vacuum molecular dynamics with a 0.25 fs time step at 300 K using
    the Langevin thermostat on 14 well-known drug molecules and two small proteins. System
    sizes range from 20 to 312 atoms. A random subsample of 128 frames from each 1ns
    trajectory is selected, and reference DFT single point calculations are performed to obtain
    QM energies and forces.

    The description copied from https://arxiv.org/pdf/1801.09319.pdf
    """

    @property
    def raw_url_name(self):
        return "ANI-MD"

    @property
    def raw_file_names(self):
        return ["ani_md_bench.h5"]


class DrugBank(COMP6Base):
    """
    DrugBank Benchmark. This benchmark is developed through a subsampling of the
    DrugBank database of real drug molecules. 837 SMILES strings containing C, N, and O
    are randomly selected. Like the GDB7to9 benchmark, the molecules are embedded in 3D
    space, structurally optimized, and normal modes are computed. DNMS is utilized to
    generate random non-equilibrium conformations.

    The description copied from https://arxiv.org/pdf/1801.09319.pdf
    """

    @property
    def raw_file_names(self):
        return ["drugbank_testset.h5"]


class GDB07to09(COMP6Base):
    """
    GDB7to9 Benchmark. The GDB-11 subsets containing 7 to 9 heavy atoms (C, N, and O)
    are subsampled and randomly embedded in 3D space using RDKit [www.rdkit.org]. A total
    of 1500 molecule SMILES [opensmiles.org] strings are selected: 500 per 7, 8, and 9 heavyatom set. The resulting structures are optimized with tight convergence criteria, and normal
    modes/force constants are computed using the reference DFT model. Finally, diverse
    normal mode sampling (DNMS) is carried out to generate non-equilibrium conformations.

    The description copied from https://arxiv.org/pdf/1801.09319.pdf
    """

    @property
    def raw_file_names(self):
        return ["gdb11_07_test500.h5", "gdb11_08_test500.h5", "gdb11_09_test500.h5"]


class GDB10to13(COMP6Base):
    """
    GDB10to13 Benchmark. Subsamples of 500 SMILES strings each from the 10 and 11
    heavy-atom subsets of GDB-11 and 1000 SMILES strings from the 12 and 13 heavyatom subsets of the GDB-13 database are randomly selected. DNMS is utilized to
    generate random non-equilibrium conformations.

    The description copied from https://arxiv.org/pdf/1801.09319.pdf
    """

    @property
    def raw_file_names(self):
        return [
            "gdb11_10_test500.h5",
            "gdb11_11_test500.h5",
            "gdb13_12_test1000.h5",
            "gdb13_13_test1000.h5",
        ]


class Tripeptides(COMP6Base):
    """
    Tripeptide Benchmark. 248 random tripeptides containing H, C, N, and O are generated
    using FASTA strings and randomly embedded in 3D space using RDKit. As with
    GDB7to9, the molecules are optimized, and normal modes are computed. DNMS is utilized
    to generate random non-equilibrium conformations.

    The description copied from https://arxiv.org/pdf/1801.09319.pdf
    """

    @property
    def raw_file_names(self):
        return ["tripeptide_full.h5"]


class S66X8(COMP6Base):
    """
    S66x8 Benchmark. This dataset is built from the original S66x850 benchmark for
    comparing accuracy between different methods in describing noncovalent interactions
    common in biological molecules. S66x8 is developed from 66 dimeric systems involving
    hydrogen bonding, pi-pi stacking, London interactions, and mixed influence interactions.
    While the keen reader might question the use of this benchmark without dispersion
    corrections, since dispersion corrections such as the D3 correction by Grimme et al. are
    a posteriori additions to the produced energy, then a comparison without the correction is
    equivalent to a comparison with the same dispersion corrections applied to both models.

    The description copied from https://arxiv.org/pdf/1801.09319.pdf
    """

    @property
    def raw_url_name(self):
        return "s66x8"

    @property
    def raw_file_names(self):
        return ["s66x8_wb97x6-31gd.h5"]


class COMP6v1(Dataset):
    """
    Superset of all COMP6 subsets (ANI-MD, DrugBank, GDB7to9, GDB10to13, Tripeptides, S66x8)
    """

    def __init__(
        self,
        root,
        transform=None,
        pre_transform=None,
        pre_filter=None,
    ):
        super().__init__(root, transform, pre_transform, pre_filter)

        self.subsets = [
            DS(root, transform, pre_transform, pre_filter)
            for DS in (ANIMD, DrugBank, GDB07to09, GDB10to13, Tripeptides, S66X8)
        ]

        self.num_samples = sum(len(subset) for subset in self.subsets)

        self.subset_indices = []
        for i_subset, subset in enumerate(self.subsets):
            for i_sample in range(len(subset)):
                self.subset_indices.append([i_subset, i_sample])
        self.subset_indices = np.array(self.subset_indices)

    def len(self):
        return self.num_samples

    def get(self, idx):
        i_subset, i_sample = self.subset_indices[idx]
        return self.subsets[i_subset][i_sample]


class COMP6v2(ANIBase):
    """Dataset for the COmprehensive Machine-learning Potential (COMP6) Benchmark Suite version 2.0

    COMP6v2 is a data set of density functional properties for molecules containing H, C, N, O, S, F, and Cl.
    It is available at different levels of theory but here we use wB97X/631Gd which was used in evaluating ANI-2x.

    References:

    - https://pubs.acs.org/doi/10.1021/acs.jctc.0c00121
    """

    # Taken from https://github.com/isayev/ASE_ANI/blob/master/ani_models/ani-2x_8x/sae_linfit.dat
    _ELEMENT_ENERGIES = {
        1: -0.5978583943827134,  # H
        6: -38.08933878049795,  # C
        7: -54.711968298621066,  # N
        8: -75.19106774742086,  # O
        9: -99.80348506781634,  # F
        16: -398.1577125334925,  # S
        17: -460.1681939421027,  # Cl
    }

    @property
    def raw_url(self):
        return "https://zenodo.org/records/10126157/files/COMP6v2_wB97X-631Gd.tar.gz"

    @property
    def raw_file_names(self):
        return [os.path.join("comp6v2_final_h5", "COMP6v2_wB97X-631Gd.h5")]

    def download(self):
        archive = download_url(self.raw_url, self.raw_dir)
        extract_tar(archive, self.raw_dir)
        os.remove(archive)

    def sample_iter(self, mol_ids=False):
        """
        In [14]: list(molecules)
        Out[14]:
        [('cm5_atomic_charges', <HDF5 dataset "cm5_atomic_charges": shape (128, 313), type "<f4">),
        ('coordinates', <HDF5 dataset "coordinates": shape (128, 312, 3), type "<f4">),
        ('energies', <HDF5 dataset "energies": shape (128,), type "<f8">),
        ('forces', <HDF5 dataset "forces": shape (128, 312, 3), type "<f4">),
        ('hirshfeld_atomic_charges', <HDF5 dataset "hirshfeld_atomic_charges": shape (128, 313), type "<f4">),
        ('hirshfeld_atomic_dipoles', <HDF5 dataset "hirshfeld_atomic_dipoles": shape (128, 313, 3), type "<f4">),
        ('species', <HDF5 dataset "species": shape (128, 312), type "<i8">)]
        """
        assert len(self.raw_paths) == 1
        with h5py.File(self.raw_paths[0]) as h5data:
            for key, data in tqdm(h5data.items(), desc="Molecule Group", leave=False):
                all_z = pt.tensor(data["species"][:], dtype=pt.long)
                all_pos = pt.tensor(data["coordinates"][:], dtype=pt.float32)
                all_y = pt.tensor(
                    data["energies"][:] * self.HARTREE_TO_EV, dtype=pt.float64
                )
                all_neg_dy = pt.tensor(
                    data["forces"][:] * self.HARTREE_TO_EV, dtype=pt.float32
                )
                n_mols = all_pos.shape[0]
                n_atoms = all_pos.shape[1]

                assert all_y.shape[0] == n_mols
                assert all_z.shape == (n_mols, n_atoms)
                assert all_pos.shape == (n_mols, n_atoms, 3)
                assert all_neg_dy.shape == (n_mols, n_atoms, 3)

                for i, (pos, y, z, neg_dy) in enumerate(
                    zip(all_pos, all_y, all_z, all_neg_dy)
                ):
                    # Create a sample
                    args = dict(z=z, pos=pos, y=y.view(1, 1), neg_dy=neg_dy)
                    if mol_ids:
                        args["mol_id"] = f"{key}_{i}"
                    data = Data(**args)

                    if data := self.filter_and_pre_transform(data):
                        yield data

    def get_atomref(self, max_z=100):
        """Atomic energy reference values for the :py:mod:`torchmdnet.priors.Atomref` prior.

        Args:
            max_z (int): Maximum atomic number

        Returns:
            torch.Tensor: Atomic energy reference values for each element in the dataset.
        """
        refs = pt.zeros(max_z)
        for key, val in self._ELEMENT_ENERGIES.items():
            refs[key] = val * self.HARTREE_TO_EV

        return refs.view(-1, 1)



# Copyright Universitat Pompeu Fabra 2020-2023  https://www.compscience.org
# Distributed under the MIT License.
# (See accompanying file README.md file or copy at http://opensource.org/licenses/MIT)

import torch
from torchmdnet.priors.base import BasePrior
from torchmdnet.models.utils import OptimizedDistance, CosineCutoff, scatter
from typing import Optional, Dict


class ZBL(BasePrior):
    """
    Implements the Ziegler-Biersack-Littmark (ZBL) potential for screened nuclear repulsion.

    This potential is described in Ziegler, J.F., Biersack, J.P., Littmark, U. "The Stopping and Range of Ions in Solids."
    (1985), specifically in equations 9 and 10 on page 147. It is an empirical potential effectively describing the
    repulsion between atoms at very short distances.

    Reference:
        Available at: https://doi.org/10.1007/978-3-642-68779-2_5

    Parameters
    ----------
    atomic_number : torch.Tensor, optional
        A 1D tensor of length max_z. `atomic_number[z]` is the atomic number of atoms with atom type z. If None, use `dataset.atomic_number`.
    distance_scale : float, optional
        Factor to multiply with coordinates stored in the dataset to convert them to meters. If None, use `dataset.distance_scale`.
    energy_scale : float, optional
        Factor to multiply with energies stored in the dataset to convert them to Joules (not J/mol). If None, use `dataset.energy_scale`.
    dataset : Dataset, optional
        Dataset object. If None, `atomic_number`, `distance_scale`, and `energy_scale` must be explicitly set.

    """

    def __init__(
        self,
        cutoff_distance,
        max_num_neighbors,
        atomic_number=None,
        distance_scale=None,
        energy_scale=None,
        dataset=None,
    ):
        super(ZBL, self).__init__()
        if atomic_number is None:
            atomic_number = dataset.atomic_number
        if distance_scale is None:
            distance_scale = dataset.distance_scale
        if energy_scale is None:
            energy_scale = dataset.energy_scale
        atomic_number = torch.as_tensor(atomic_number, dtype=torch.long)
        self.register_buffer("atomic_number", atomic_number)
        self.distance = OptimizedDistance(
            0, cutoff_distance, max_num_pairs=-max_num_neighbors
        )
        self.cutoff = CosineCutoff(cutoff_upper=cutoff_distance)
        self.cutoff_distance = cutoff_distance
        self.max_num_neighbors = max_num_neighbors
        self.distance_scale = float(distance_scale)
        self.energy_scale = float(energy_scale)

    def get_init_args(self):
        return {
            "cutoff_distance": self.cutoff_distance,
            "max_num_neighbors": self.max_num_neighbors,
            "atomic_number": self.atomic_number.tolist(),
            "distance_scale": self.distance_scale,
            "energy_scale": self.energy_scale,
        }

    def reset_parameters(self):
        pass

    def post_reduce(
        self,
        y,
        z,
        pos,
        batch,
        box: Optional[torch.Tensor] = None,
        extra_args: Optional[Dict[str, torch.Tensor]] = None,
    ):
        edge_index, distance, _ = self.distance(pos, batch, box)
        if edge_index.shape[1] == 0:
            return y
        atomic_number = self.atomic_number[z[edge_index]]
        # 5.29e-11 is the Bohr radius in meters.  All other numbers are magic constants from the ZBL potential.
        a = (
            0.8854
            * 5.29177210903e-11
            / (atomic_number[0] ** 0.23 + atomic_number[1] ** 0.23)
        )
        d = distance * self.distance_scale / a
        f = (
            0.1818 * torch.exp(-3.2 * d)
            + 0.5099 * torch.exp(-0.9423 * d)
            + 0.2802 * torch.exp(-0.4029 * d)
            + 0.02817 * torch.exp(-0.2016 * d)
        )
        f *= self.cutoff(distance)
        # Compute the energy, converting to the dataset's units.  Multiply by 0.5 because every atom pair
        # appears twice.
        energy = f * atomic_number[0] * atomic_number[1] / distance
        energy = (
            0.5
            * (2.30707755e-28 / self.energy_scale / self.distance_scale)
            * scatter(energy, batch[edge_index[0]], dim=0, reduce="sum")
        )
        if energy.shape[0] < y.shape[0]:
            energy = torch.nn.functional.pad(energy, (0, y.shape[0] - energy.shape[0]))
        energy = energy.reshape(y.shape)
        return y + energy



# Copyright Universitat Pompeu Fabra 2020-2023  https://www.compscience.org
# Distributed under the MIT License.
# (See accompanying file README.md file or copy at http://opensource.org/licenses/MIT)

from torchmdnet.priors.base import BasePrior
from typing import Optional, Dict
import torch
from torch import nn, Tensor
from lightning_utilities.core.rank_zero import rank_zero_warn


class Atomref(BasePrior):
    r"""Atomref prior model.

    This prior model is used to add atomic reference values to the input features. The atomic reference values are stored in an embedding layer and are added to the input features as:

    .. math::

        x' = x + \\textrm{atomref}(z)

    where :math:`x` is the input feature tensor, :math:`z` is the atomic number tensor, and :math:`\\textrm{atomref}` is the embedding layer. The atomic reference values are stored in the embedding layer and can be trainable.

    When using this in combination with some dataset, the dataset class must implement the function `get_atomref`, which returns the atomic reference values as a tensor.

    Args:
        max_z (int, optional): Maximum atomic number to consider. If `dataset` is not `None`, this argument is ignored.
        dataset (torch_geometric.data.Dataset, optional): A dataset from which to extract the atomref values.
        trainable (bool, optional): If `False`, the atomref values are not trainable. (default: `False`)
        enable (bool, optional): If `False`, the prior is disabled. This is useful if you want to add the reference energies only during inference (or training) (default: `True`)
    """

    def __init__(self, max_z=None, dataset=None, trainable=False, enable=True):
        super().__init__()
        if max_z is None and dataset is None:
            raise ValueError("Can't instantiate Atomref prior, all arguments are None.")
        if dataset is None:
            assert max_z is not None, "max_z must be provided if dataset is None."
            atomref = torch.zeros(max_z, 1)
        else:
            atomref = dataset.get_atomref()
            if atomref is None:
                rank_zero_warn(
                    "The atomref returned by the dataset is None, defaulting to zeros with max. "
                    "atomic number 99. Maybe atomref is not defined for the current target."
                )
                atomref = torch.zeros(100, 1)

        if atomref.ndim == 1:
            atomref = atomref.view(-1, 1)
        self.register_buffer("initial_atomref", atomref)
        self.atomref = nn.Embedding(
            len(atomref), 1, _freeze=not trainable, _weight=atomref
        )
        self.enable = enable

    def reset_parameters(self):
        self.atomref.weight.data.copy_(self.initial_atomref)

    def get_init_args(self):
        return dict(
            max_z=self.initial_atomref.size(0),
            trainable=self.atomref.weight.requires_grad,
            enable=self.enable,
        )

    def pre_reduce(
        self,
        x: Tensor,
        z: Tensor,
        pos: Tensor,
        batch: Optional[Tensor] = None,
        extra_args: Optional[Dict[str, Tensor]] = None,
    ):
        """Applies the stored atomref to the input as:

        .. math::

            x' = x + \\textrm{atomref}(z)

        .. note:: The atomref operation is an embedding lookup that can be trainable if the `trainable` argument is set to `False`.

        .. note:: This call becomes a no-op if the `enable` argument is set to `False`.

        Args:
            x (Tensor): Input feature tensor.
            z (Tensor): Atomic number tensor.
            pos (Tensor): Atomic positions tensor. Unused.
            batch (Tensor, optional): Batch tensor. Unused. (default: `None`).
            extra_args (Dict[str, Tensor], optional): Extra arguments. Unused. (default: `None`)


        """
        if self.enable:
            return x + self.atomref(z)
        else:
            return x


class LearnableAtomref(Atomref):
    r"""LearnableAtomref prior model.

    This prior model is used to add learned atomic reference values to the input features. The atomic reference values are learned as an embedding layer and are added to the input features as:

    .. math::

        x' = x + \\textrm{atomref}(z)

    where :math:`x` is the input feature tensor, :math:`z` is the atomic number tensor, and :math:`\\textrm{atomref}` is the embedding layer.


    Args:
        max_z (int, optional): Maximum atomic number to consider.
    """

    def __init__(self, max_z=None):
        super().__init__(max_z, trainable=True, enable=True)



# Copyright Universitat Pompeu Fabra 2020-2023  https://www.compscience.org
# Distributed under the MIT License.
# (See accompanying file README.md file or copy at http://opensource.org/licenses/MIT)

from abc import abstractmethod, ABCMeta
from typing import Optional
import torch
from torch import nn
from torchmdnet.models.utils import (
    GatedEquivariantBlock,
    scatter,
    MLP,
    OptimizedDistance,
)
from torchmdnet.utils import atomic_masses
from warnings import warn
import numpy as np

__all__ = ["Scalar", "DipoleMoment", "ElectronicSpatialExtent"]


class OutputModel(nn.Module, metaclass=ABCMeta):
    """Base class for output models.

    Derive this class to make custom output models.
    As an example, have a look at the :py:mod:`torchmdnet.output_modules.Scalar` output model.
    """

    def __init__(self, allow_prior_model, reduce_op, static_shapes=False):
        super(OutputModel, self).__init__()
        self.allow_prior_model = allow_prior_model
        self.reduce_op = reduce_op
        self.static_shapes = static_shapes
        self.dim_size = 0

    def reset_parameters(self):
        pass

    @abstractmethod
    def pre_reduce(self, x, v, z, pos, batch, box: Optional[torch.Tensor] = None):
        return

    def reduce(self, x, batch, num_systems: Optional[int] = None):
        # If num_systems is explicitly provided, use it directly.
        # This avoids the .item() call that is incompatible with torch.export tracing,
        # and allows dynamic atom counts without a warm-up step.
        if num_systems is not None:
            return scatter(x, batch, dim=0, dim_size=num_systems, reduce=self.reduce_op)
        # torch.compile and torch.export don't support .item() calls during tracing
        # The model should be warmed up before compilation to set the correct dim_size
        if torch.compiler.is_compiling():
            pass
        elif torch.jit.is_scripting():
            # TorchScript doesn't support torch.cuda.is_current_stream_capturing()
            # For CPU, always update dim_size (no CUDA graphs on CPU)
            # For CUDA with static_shapes, only update once (first call sets dim_size for CUDA graph capture)
            # For CUDA without static_shapes, always update (dynamic batch sizes)
            if not x.is_cuda or not self.static_shapes or self.dim_size == 0:
                self.dim_size = int(batch.max().item() + 1)
        else:
            is_capturing = x.is_cuda and torch.cuda.is_current_stream_capturing()
            if not x.is_cuda or not is_capturing:
                self.dim_size = int(batch.max().item() + 1)
            if is_capturing:
                assert (
                    self.dim_size > 0
                ), "Warming up is needed before capturing the model into a CUDA graph"
                warn(
                    "CUDA graph capture will lock the batch to the current number of samples ({}). Changing this will result in a crash".format(
                        self.dim_size
                    )
                )
        return scatter(x, batch, dim=0, dim_size=self.dim_size, reduce=self.reduce_op)

    def post_reduce(self, x):
        return x


class Scalar(OutputModel):
    def __init__(
        self,
        hidden_channels,
        activation="silu",
        allow_prior_model=True,
        reduce_op="sum",
        dtype=torch.float,
        static_shapes=False,
        **kwargs,
    ):
        super(Scalar, self).__init__(
            allow_prior_model=allow_prior_model,
            reduce_op=reduce_op,
            static_shapes=static_shapes,
        )
        self.output_network = MLP(
            in_channels=hidden_channels,
            out_channels=1,
            hidden_channels=hidden_channels // 2,
            activation=activation,
            num_hidden_layers=kwargs.get("num_layers", 0),
            dtype=dtype,
        )
        self.reset_parameters()

    def reset_parameters(self):
        self.output_network.reset_parameters()

    def pre_reduce(
        self,
        x,
        v: Optional[torch.Tensor],
        z,
        pos,
        batch,
        box: Optional[torch.Tensor] = None,
    ):
        return self.output_network(x)


class EquivariantScalar(OutputModel):
    def __init__(
        self,
        hidden_channels,
        activation="silu",
        allow_prior_model=True,
        reduce_op="sum",
        dtype=torch.float,
        static_shapes=False,
        **kwargs,
    ):
        super(EquivariantScalar, self).__init__(
            allow_prior_model=allow_prior_model,
            reduce_op=reduce_op,
            static_shapes=static_shapes,
        )
        if kwargs.get("num_layers", 0) > 0:
            warn("num_layers is not used in EquivariantScalar")
        self.output_network = nn.ModuleList(
            [
                GatedEquivariantBlock(
                    hidden_channels,
                    hidden_channels // 2,
                    activation=activation,
                    scalar_activation=True,
                    dtype=dtype,
                ),
                GatedEquivariantBlock(
                    hidden_channels // 2, 1, activation=activation, dtype=dtype
                ),
            ]
        )

        self.reset_parameters()

    def reset_parameters(self):
        for layer in self.output_network:
            layer.reset_parameters()

    def pre_reduce(self, x, v, z, pos, batch, box: Optional[torch.Tensor] = None):
        for layer in self.output_network:
            x, v = layer(x, v)
        # include v in output to make sure all parameters have a gradient
        return x + v.sum() * 0


class DipoleMoment(Scalar):
    def __init__(
        self,
        hidden_channels,
        activation="silu",
        reduce_op="sum",
        dtype=torch.float,
        static_shapes=False,
        **kwargs,
    ):
        super(DipoleMoment, self).__init__(
            hidden_channels,
            activation,
            allow_prior_model=False,
            reduce_op=reduce_op,
            dtype=dtype,
            static_shapes=static_shapes,
            **kwargs,
        )
        atomic_mass = torch.from_numpy(atomic_masses).to(dtype)
        self.register_buffer("atomic_mass", atomic_mass)

    def pre_reduce(
        self,
        x,
        v: Optional[torch.Tensor],
        z,
        pos,
        batch,
        box: Optional[torch.Tensor] = None,
    ):
        x = self.output_network(x)

        # Get center of mass.
        mass = self.atomic_mass[z].view(-1, 1)
        c = scatter(mass * pos, batch, dim=0) / scatter(mass, batch, dim=0)
        x = x * (pos - c[batch])
        return x

    def post_reduce(self, x):
        return torch.norm(x, dim=-1, keepdim=True)


class EquivariantDipoleMoment(EquivariantScalar):
    def __init__(
        self,
        hidden_channels,
        activation="silu",
        reduce_op="sum",
        dtype=torch.float,
        static_shapes=False,
        **kwargs,
    ):
        super(EquivariantDipoleMoment, self).__init__(
            hidden_channels,
            activation,
            allow_prior_model=False,
            reduce_op=reduce_op,
            dtype=dtype,
            static_shapes=static_shapes,
            **kwargs,
        )
        atomic_mass = torch.from_numpy(atomic_masses).to(dtype)
        self.register_buffer("atomic_mass", atomic_mass)

    def pre_reduce(self, x, v, z, pos, batch, box: Optional[torch.Tensor] = None):
        for layer in self.output_network:
            x, v = layer(x, v)

        # Get center of mass.
        mass = self.atomic_mass[z].view(-1, 1)
        c = scatter(mass * pos, batch, dim=0) / scatter(mass, batch, dim=0)
        x = x * (pos - c[batch])
        return x + v.squeeze()

    def post_reduce(self, x):
        return torch.norm(x, dim=-1, keepdim=True)


class ElectronicSpatialExtent(OutputModel):
    def __init__(
        self,
        hidden_channels,
        activation="silu",
        reduce_op="sum",
        dtype=torch.float,
        static_shapes=False,
        **kwargs,
    ):
        super(ElectronicSpatialExtent, self).__init__(
            allow_prior_model=False, reduce_op=reduce_op, static_shapes=static_shapes
        )
        self.output_network = MLP(
            in_channels=hidden_channels,
            out_channels=1,
            hidden_channels=hidden_channels // 2,
            activation=activation,
            num_hidden_layers=kwargs.get("num_layers", 0),
            dtype=dtype,
        )
        atomic_mass = torch.from_numpy(atomic_masses).to(dtype)
        self.register_buffer("atomic_mass", atomic_mass)

        self.reset_parameters()

    def reset_parameters(self):
        self.output_network.reset_parameters()

    def pre_reduce(
        self,
        x,
        v: Optional[torch.Tensor],
        z,
        pos,
        batch,
        box: Optional[torch.Tensor] = None,
    ):
        x = self.output_network(x)

        # Get center of mass.
        mass = self.atomic_mass[z].view(-1, 1)
        c = scatter(mass * pos, batch, dim=0) / scatter(mass, batch, dim=0)

        x = torch.norm(pos - c[batch], dim=1, keepdim=True) ** 2 * x
        return x


class EquivariantElectronicSpatialExtent(ElectronicSpatialExtent):
    pass


class EquivariantVectorOutput(EquivariantScalar):
    def __init__(
        self,
        hidden_channels,
        activation="silu",
        reduce_op="sum",
        dtype=torch.float,
        static_shapes=False,
        **kwargs,
    ):
        super(EquivariantVectorOutput, self).__init__(
            hidden_channels,
            activation,
            allow_prior_model=False,
            reduce_op="sum",
            dtype=dtype,
            static_shapes=static_shapes,
            **kwargs,
        )

    def pre_reduce(self, x, v, z, pos, batch, box: Optional[torch.Tensor] = None):
        for layer in self.output_network:
            x, v = layer(x, v)
        return v.squeeze()


def _exp_cutoff(d, rc: float):
    """short range cutoff to damp close range coulomb interaction

    Implementation copied form aimnetcentral
    """
    fc = (
        torch.exp(-1.0 / (1.0 - (d / rc).clamp(0, 1.0 - 1e-6).pow(2)))
        / 0.36787944117144233
    )
    return fc


def _triu_indices(x, batch):
    N = x.shape[0]
    i, j = torch.triu_indices(N, N, 1, device=x.device, dtype=torch.long).unbind(0)
    mask = batch[i] == batch[j]
    i = i[mask]
    j = j[mask]
    return torch.stack([i, j])


class ScalarPlusWeightedCoulomb(OutputModel):
    """normal scalar output plus a coulomb interaction using predicted charges

    It assumes the input features are shape [N, hidden_channels + q_channels]
    where q_channels are the predicted partial charges (can be 1 or length N)
    If more than 1 the coulomb interaction is computed seperately for each charge
    channel and then a weighted mean is taken of the resulting energies.

    Implementation modifed from aimnetcentral
    """

    def __init__(
        self,
        hidden_channels,
        activation="silu",
        allow_prior_model=True,
        reduce_op="sum",
        dtype=torch.float,
        static_shapes=False,
        cutoff: float | None = None,
        **kwargs,
    ):
        self.static_shapes = static_shapes
        super(ScalarPlusWeightedCoulomb, self).__init__(
            allow_prior_model=allow_prior_model,
            reduce_op=reduce_op,
            static_shapes=static_shapes,
        )
        self.hidden_channels = hidden_channels
        self.output_network = MLP(
            in_channels=hidden_channels,
            out_channels=1,
            hidden_channels=hidden_channels // 2,
            activation=activation,
            num_hidden_layers=kwargs.get("num_hidden_layers", 0),
            dtype=dtype,
        )

        self.q_dim = kwargs["q_dim"]
        self.num_interaction_layers = kwargs["num_layers"]
        self.layer_weights = kwargs["q_weights"]
        self.cutoff = kwargs["coulomb_cutoff"]

        assert len(self.layer_weights) == self.num_interaction_layers + 1

        _w = torch.zeros((self.num_interaction_layers + 1, self.q_dim), dtype=dtype)
        for i in range(self.num_interaction_layers + 1):
            _w[i, :] = torch.tensor(self.layer_weights[i], dtype=dtype)

        _w = _w.flatten()

        self.register_buffer("qweights", _w)

        # from ase.units from aimnet2
        _Hartree = 27.211386024367243
        _half_Hartree = 0.5 * _Hartree
        _Bohr = 0.5291772105638411
        self._factor = _half_Hartree * _Bohr

        if self.cutoff is None:
            # no cutoff we do N^2 all-to-all, e.g. small mol in vacuum
            self.mode = "all_to_all"
            self.distance = None

        else:
            # we use a cutoff and Reaction Field method
            # this will work with PBC if the box is larger than 2 * cutoff
            self.mode = "cutoff"

            if (
                "coulomb_max_num_neighbors" in kwargs
                and kwargs["coulomb_max_num_neighbors"] is not None
            ):
                max_num_neighbors = kwargs["coulomb_max_num_neighbors"]
            else:
                # if not provided we need to set max_num_pairs sensibly
                # we assume uniform water like density
                _density = 0.1  # atoms per cubic Angstrom
                _volume = 4 / 3 * np.pi * self.cutoff**3
                max_num_neighbors = int(_density * _volume)

            self.distance = OptimizedDistance(
                cutoff_lower=0.0,
                cutoff_upper=self.cutoff,
                max_num_pairs=-max_num_neighbors,
                return_vecs=False,
                resize_to_fit=not self.static_shapes,
                include_transpose=False,
                loop=False,
                strategy=kwargs.get("coulomb_neighbor_strategy", "brute"),
            )
            self.epsilon_solvent = kwargs.get("coulomb_epsilon_solvent", 78.3)

        # for torchscript
        self.edge_index = torch.empty(2, 0)

        self.reset_parameters()

    def reset_parameters(self):
        self.output_network.reset_parameters()

    def pre_reduce(
        self,
        x,
        v: Optional[torch.Tensor],
        z,
        pos,
        batch,
        box: Optional[torch.Tensor] = None,
    ):

        if box is not None:
            if self.mode == "all_to_all":
                raise ValueError("PBC is not supported with coulomb_cutoff = None")

        # we assume the charges have been added at the end of x!
        charges = x[:, self.hidden_channels :]

        if not self.static_shapes:  # check we have the expected number of charges
            assert charges.shape[1] == self.qweights.shape[0]

        x = x[:, : self.hidden_channels]
        x = self.output_network(x)  # energy per atom from main MLPs

        if self.mode == "all_to_all":
            # N^2 code path

            # we have to some some warmup to lock in the batch size if we want to script or compile with fullgraph=True

            # torch.compile and torch.export don't support .item() calls during tracing
            # The model should be warmed up before compilation to set the correct dim_size
            if torch.compiler.is_compiling():
                assert self.edge_index.shape[1] > 0

            elif torch.jit.is_scripting():
                # TorchScript doesn't support torch.cuda.is_current_stream_capturing()
                # For CPU, always update dim_size (no CUDA graphs on CPU)
                # For CUDA with static_shapes, only update once (first call sets dim_size for CUDA graph capture)
                # For CUDA without static_shapes, always update (dynamic batch sizes)
                if (
                    not x.is_cuda
                    or not self.static_shapes
                    or self.edge_index.shape[1] == 0
                ):
                    self.edge_index = _triu_indices(x, batch)
            else:
                is_capturing = x.is_cuda and torch.cuda.is_current_stream_capturing()
                if not x.is_cuda or not is_capturing:
                    self.edge_index = _triu_indices(x, batch)

                if is_capturing:
                    assert (
                        self.edge_index is not None
                    ), "Warming up is needed before capturing the model into a CUDA graph"
                    warn(
                        "CUDA graph capture will lock the batch to the current number of samples ({}). Changing this will result in a crash".format(
                            self.dim_size
                        )
                    )

            q_i = charges[self.edge_index[0]]
            q_j = charges[self.edge_index[1]]

            d_ij = torch.linalg.norm(
                pos[self.edge_index[0]] - pos[self.edge_index[1]], dim=-1
            )

            # charge product
            q_ij = q_i * q_j

            # short range damping, hyperparameter is from aimnet2
            fc = 1.0 - _exp_cutoff(d_ij, 4.6)

            # per edge Coulomb energy with short range damping
            e_ij = fc.unsqueeze(-1) * q_ij / d_ij.unsqueeze(-1)

            # scale with Coulomb constant in eV and Angstrom unit system!
            e_ij = self._factor * e_ij

            # now do the weighted mean over features
            e_ij = torch.sum(e_ij * self.qweights.unsqueeze(0), dim=-1) / torch.sum(
                self.qweights
            )

            # sum into nodes
            e_i = torch.zeros(x.shape[0], device=x.device, dtype=x.dtype)
            e_i = e_i.index_add(0, self.edge_index[0], e_ij)
            e_i = e_i.index_add(0, self.edge_index[1], e_ij)

        else:
            # cutoff code path
            edge_index, edge_weight, _ = self.distance(pos, batch, box=box)

            if self.static_shapes:
                # -1 in edge index need to be mapped to an extra dummy atom
                mask = (edge_index[0] < 0).unsqueeze(0).expand_as(edge_index)
                edge_index = edge_index.masked_fill(mask, z.shape[0])
                edge_weight = edge_weight.masked_fill(mask[0], 0)

                # we need a dummy atom
                charges = torch.cat(
                    [
                        charges,
                        torch.zeros(
                            1,
                            charges.shape[1],
                            device=charges.device,
                            dtype=charges.dtype,
                        ),
                    ],
                    dim=0,
                )

            q_i = charges[edge_index[0]]
            q_j = charges[edge_index[1]]

            d_ij = edge_weight

            # charge product
            q_ij = q_i * q_j

            # short range damping, hyperparameter is from aimnet2
            fc = 1.0 - _exp_cutoff(d_ij, 4.6)

            # reaction field implementation
            # uses the OpenMM version
            k_rf = (
                (1.0 / self.cutoff**3)
                * (self.epsilon_solvent - 1.0)
                / (2.0 * self.epsilon_solvent + 1.0)
            )
            c_rf = (
                (1.0 / self.cutoff)
                * (3.0 * self.epsilon_solvent)
                / (2.0 * self.epsilon_solvent + 1.0)
            )
            e_ij = (
                fc.unsqueeze(-1)
                * q_ij
                * (1.0 / d_ij + k_rf * d_ij**2 - c_rf).unsqueeze(-1)
            )

            # scale with Coulomb constant in eV and Angstrom unit system!
            e_ij = self._factor * e_ij

            # now do the weighted mean over features
            e_ij = torch.sum(e_ij * self.qweights.unsqueeze(0), dim=-1) / torch.sum(
                self.qweights
            )

            # sum into nodes
            e_i = torch.zeros(
                x.shape[0] + 1 if self.static_shapes else x.shape[0],
                device=x.device,
                dtype=x.dtype,
            )
            e_i = e_i.index_add(0, edge_index[0], e_ij)
            e_i = e_i.index_add(0, edge_index[1], e_ij)

            if self.static_shapes:
                e_i = e_i[:-1]

        # add to normal per atom output with correct shape
        x = x + e_i.unsqueeze(-1)

        return x



# Copyright Universitat Pompeu Fabra 2020-2023  https://www.compscience.org
# Distributed under the MIT License.
# (See accompanying file README.md file or copy at http://opensource.org/licenses/MIT)

import h5py
import numpy as np
import os
import torch as pt
from torch_geometric.data import Data, download_url, extract_tar
from torchmdnet.datasets.memdataset import MemmappedDataset
from tqdm import tqdm
import warnings


class ANIBase(MemmappedDataset):
    """ANI Dataset Classes

    A foundational dataset class for handling the ANI datasets. ANI (ANAKIN-ME or Accurate NeurAl networK engINe for Molecular Energies)
    is a deep learning method trained on quantum mechanical DFT calculations to predict accurate and transferable potentials for organic molecules.

    Key features of ANI:

    - Utilizes a modified version of the Behler and Parrinello symmetry functions to construct single-atom atomic environment vectors (AEV) for molecular representation.
    - AEVs enable the training of neural networks over both configurational and conformational space.
    - The ANI-1 potential was trained on a subset of the GDB databases with up to 8 heavy atoms.
    - ANI-1x and ANI-1ccx datasets provide diverse quantum mechanical properties for organic molecules:
       -  ANI-1x contains multiple QM properties from 5M density functional theory calculations.
       -  ANI-1ccx contains 500k data points obtained with an accurate CCSD(T)/CBS extrapolation.
    - Properties include energies, atomic forces, multipole moments, atomic charges, and more for the chemical elements C, H, N, and O.
    - Developed through active learning, an automated data diversification process.

    References:

    - Smith, J. S., Isayev, O., & Roitberg, A. E. (2017). ANI-1: an extensible neural network potential with DFT accuracy at force field computational cost. Chemical Science, 8(4), 3192-3203.
    - Smith, J. S., Zubatyuk, R., Nebgen, B., Lubbers, N., Barros, K., Roitberg, A. E., Isayev, O., & Tretiak, S. (2020). The ANI-1ccx and ANI-1x data sets, coupled-cluster and density functional theory properties for molecules. Scientific Data, 7, Article 134.
    """

    HARTREE_TO_EV = 27.211386246  #::meta private:

    @property
    def raw_url(self):
        raise NotImplementedError

    @property
    def raw_file_names(self):
        raise NotImplementedError

    def get_atomref(self, max_z=100):
        """Atomic energy reference values for the :py:mod:`torchmdnet.priors.Atomref` prior.

        Args:
            max_z (int): Maximum atomic number

        Returns:
            torch.Tensor: Atomic energy reference values for each element in the dataset.
        """
        refs = pt.zeros(max_z)
        for key, val in self._ELEMENT_ENERGIES.items():
            refs[key] = val * self.HARTREE_TO_EV

        return refs.view(-1, 1)

    def __init__(
        self,
        root,
        transform=None,
        pre_transform=None,
        pre_filter=None,
        properties=("y", "neg_dy"),
    ):
        self.name = self.__class__.__name__
        super().__init__(
            root,
            transform,
            pre_transform,
            pre_filter,
            properties=properties,
        )

    def filter_and_pre_transform(self, data):
        if self.pre_filter is not None and not self.pre_filter(data):
            return None

        if self.pre_transform is not None:
            data = self.pre_transform(data)

        return data


class ANI1(ANIBase):
    __doc__ = ANIBase.__doc__
    _ELEMENT_ENERGIES = {
        1: -0.500607632585,
        6: -37.8302333826,
        7: -54.5680045287,
        8: -75.0362229210,
    }

    def __init__(
        self,
        root,
        transform=None,
        pre_transform=None,
        pre_filter=None,
    ):
        self.name = self.__class__.__name__
        super().__init__(
            root,
            transform,
            pre_transform,
            pre_filter,
            properties=("y",),
        )

    @property
    def raw_url(self):
        return "https://ndownloader.figshare.com/files/9057631"

    @property
    def raw_file_names(self):
        return [
            os.path.join("ANI-1_release", f"ani_gdb_s{i:02d}.h5") for i in range(1, 9)
        ]

    def download(self):
        archive = download_url(self.raw_url, self.raw_dir)
        extract_tar(archive, self.raw_dir)
        os.remove(archive)

    def sample_iter(self, mol_ids=False):
        atomic_numbers = {b"H": 1, b"C": 6, b"N": 7, b"O": 8}

        for path in tqdm(self.raw_paths, desc="Files"):
            molecules = list(h5py.File(path).values())[0].items()

            for mol_id, mol in tqdm(molecules, desc="Molecules", leave=False):
                z = pt.tensor(
                    [atomic_numbers[atom] for atom in mol["species"]], dtype=pt.long
                )
                all_pos = pt.tensor(mol["coordinates"][:], dtype=pt.float32)
                all_y = pt.tensor(
                    mol["energies"][:] * self.HARTREE_TO_EV, dtype=pt.float64
                )

                assert all_pos.shape[0] == all_y.shape[0]
                assert all_pos.shape[1] == z.shape[0]
                assert all_pos.shape[2] == 3

                for pos, y in zip(all_pos, all_y):
                    # Create a sample
                    args = dict(z=z, pos=pos, y=y.view(1, 1))
                    if mol_ids:
                        args["mol_id"] = mol_id
                    data = Data(**args)

                    if data := self.filter_and_pre_transform(data):
                        yield data


class ANI1XBase(ANIBase):
    @property
    def raw_url(self):
        return "https://figshare.com/ndownloader/files/18112775"

    @property
    def raw_file_names(self):
        return "ani1x-release.h5"

    def download(self):
        file = download_url(self.raw_url, self.raw_dir)
        assert len(self.raw_paths) == 1
        os.rename(file, self.raw_paths[0])


class ANI1X(ANI1XBase):
    __doc__ = ANIBase.__doc__
    _ELEMENT_ENERGIES = {
        1: -0.600952980000,
        6: -38.08316124000,
        7: -54.70775770000,
        8: -75.19446356000,
    }

    def sample_iter(self, mol_ids=False):
        assert len(self.raw_paths) == 1

        with h5py.File(self.raw_paths[0]) as h5:
            for mol_id, mol in tqdm(h5.items(), desc="Molecules"):
                z = pt.tensor(mol["atomic_numbers"][:], dtype=pt.long)
                all_pos = pt.tensor(mol["coordinates"][:], dtype=pt.float32)
                all_y = pt.tensor(
                    mol["wb97x_dz.energy"][:] * self.HARTREE_TO_EV, dtype=pt.float64
                )
                all_neg_dy = pt.tensor(
                    mol["wb97x_dz.forces"][:] * self.HARTREE_TO_EV, dtype=pt.float32
                )

                assert all_pos.shape[0] == all_y.shape[0]
                assert all_pos.shape[1] == z.shape[0]
                assert all_pos.shape[2] == 3

                assert all_neg_dy.shape[0] == all_y.shape[0]
                assert all_neg_dy.shape[1] == z.shape[0]
                assert all_neg_dy.shape[2] == 3

                for pos, y, neg_dy in zip(all_pos, all_y, all_neg_dy):
                    if y.isnan() or neg_dy.isnan().any():
                        continue

                    # Create a sample
                    args = dict(z=z, pos=pos, y=y.view(1, 1), neg_dy=neg_dy)
                    if mol_ids:
                        args["mol_id"] = mol_id
                    data = Data(**args)

                    if data := self.filter_and_pre_transform(data):
                        yield data


class ANI1CCX(ANI1XBase):
    __doc__ = ANIBase.__doc__
    _ELEMENT_ENERGIES = {
        1: -0.5991501324919538,
        6: -38.03750806057356,
        7: -54.67448347695333,
        8: -75.16043537275567,
    }

    def __init__(
        self,
        root,
        transform=None,
        pre_transform=None,
        pre_filter=None,
    ):
        self.name = self.__class__.__name__
        super().__init__(
            root,
            transform,
            pre_transform,
            pre_filter,
            properties=("y",),
        )

    def sample_iter(self, mol_ids=False):
        assert len(self.raw_paths) == 1

        with h5py.File(self.raw_paths[0]) as h5:
            for mol_id, mol in tqdm(h5.items(), desc="Molecules"):
                z = pt.tensor(mol["atomic_numbers"][:], dtype=pt.long)
                all_pos = pt.tensor(mol["coordinates"][:], dtype=pt.float32)
                all_y = pt.tensor(
                    mol["ccsd(t)_cbs.energy"][:] * self.HARTREE_TO_EV, dtype=pt.float64
                )

                assert all_pos.shape[0] == all_y.shape[0]
                assert all_pos.shape[1] == z.shape[0]
                assert all_pos.shape[2] == 3

                for pos, y in zip(all_pos, all_y):
                    if y.isnan():
                        continue

                    # Create a sample
                    args = dict(z=z, pos=pos, y=y.view(1, 1))
                    if mol_ids:
                        args["mol_id"] = mol_id
                    data = Data(**args)

                    if data := self.filter_and_pre_transform(data):
                        yield data


class ANI2X(ANIBase):
    __doc__ = ANIBase.__doc__

    # Taken from https://github.com/isayev/ASE_ANI/blob/master/ani_models/ani-2x_8x/sae_linfit.dat
    _ELEMENT_ENERGIES = {
        1: -0.5978583943827134,  # H
        6: -38.08933878049795,  # C
        7: -54.711968298621066,  # N
        8: -75.19106774742086,  # O
        9: -99.80348506781634,  # F
        16: -398.1577125334925,  # S
        17: -460.1681939421027,  # Cl
    }

    @property
    def raw_url(self):
        return "https://zenodo.org/records/10108942/files/ANI-2x-wB97X-631Gd.tar.gz"

    @property
    def raw_file_names(self):
        return [os.path.join("final_h5", "ANI-2x-wB97X-631Gd.h5")]

    def download(self):
        archive = download_url(self.raw_url, self.raw_dir)
        extract_tar(archive, self.raw_dir)
        os.remove(archive)

    def sample_iter(self, mol_ids=False):
        """
        In [15]: list(molecules)
        Out[15]:
        [('coordinates', <HDF5 dataset "coordinates": shape (5706, 2, 3), type "<f4">),
        ('energies', <HDF5 dataset "energies": shape (5706,), type "<f8">),
        ('forces', <HDF5 dataset "forces": shape (5706, 2, 3), type "<f8">),
        ('species', <HDF5 dataset "species": shape (5706, 2), type "<i8">)]
        """
        assert len(self.raw_paths) == 1
        with h5py.File(self.raw_paths[0]) as h5data:
            for key, data in tqdm(h5data.items(), desc="Molecule Group", leave=False):
                all_z = pt.tensor(data["species"][:], dtype=pt.long)
                all_pos = pt.tensor(data["coordinates"][:], dtype=pt.float32)
                all_y = pt.tensor(
                    data["energies"][:] * self.HARTREE_TO_EV, dtype=pt.float64
                )
                all_neg_dy = pt.tensor(
                    data["forces"][:] * self.HARTREE_TO_EV, dtype=pt.float32
                )
                n_mols = all_pos.shape[0]
                n_atoms = all_pos.shape[1]

                assert all_y.shape[0] == n_mols
                assert all_z.shape == (n_mols, n_atoms)
                assert all_pos.shape == (n_mols, n_atoms, 3)
                assert all_neg_dy.shape == (n_mols, n_atoms, 3)

                for i, (pos, y, z, neg_dy) in enumerate(
                    zip(all_pos, all_y, all_z, all_neg_dy)
                ):
                    # Create a sample
                    args = dict(z=z, pos=pos, y=y.view(1, 1), neg_dy=neg_dy)
                    if mol_ids:
                        args["mol_id"] = f"{key}_{i}"
                    data = Data(**args)

                    if data := self.filter_and_pre_transform(data):
                        yield data



# Copyright Universitat Pompeu Fabra 2020-2023  https://www.compscience.org
# Distributed under the MIT License.
# (See accompanying file README.md file or copy at http://opensource.org/licenses/MIT)

import sys
import os
import yaml
import argparse
import logging
import torch
import lightning.pytorch as pl
from lightning.pytorch.loggers import WandbLogger, CSVLogger, TensorBoardLogger
from lightning.pytorch.callbacks import (
    ModelCheckpoint,
    EarlyStopping,
)
from torchmdnet.module import LNNP
from torchmdnet import datasets, priors, models
from torchmdnet.data import DataModule
from torchmdnet.loss import loss_class_mapping
from torchmdnet.models import output_modules
from torchmdnet.models.model import create_prior_models
from torchmdnet.models.utils import rbf_class_mapping, act_class_mapping, dtype_mapping
from torchmdnet.utils import (
    LoadFromFile,
    LoadFromCheckpoint,
    save_argparse,
    number,
    check_logs,
)
from lightning_utilities.core.rank_zero import rank_zero_warn


def get_argparse():
    # fmt: off
    parser = argparse.ArgumentParser(description='Training')
    parser.add_argument('--load-model', action=LoadFromCheckpoint, help='Restart training using a model checkpoint')  # keep first
    parser.add_argument('--conf', '-c', type=open, action=LoadFromFile, help='Configuration yaml file')  # keep second
    parser.add_argument('--num-epochs', default=300, type=int, help='number of epochs')
    parser.add_argument('--batch-size', default=32, type=int, help='batch size')
    parser.add_argument('--inference-batch-size', default=None, type=int, help='Batchsize for validation and tests.')
    parser.add_argument('--lr', default=1e-4, type=float, help='learning rate')
    parser.add_argument('--lr-patience', type=int, default=10, help='Patience for lr-schedule. Patience per eval-interval of validation')
    parser.add_argument('--lr-metric', type=str, default='val', choices=['train', 'val'], help='Metric to monitor when deciding whether to reduce learning rate')
    parser.add_argument('--lr-min', type=float, default=1e-6, help='Minimum learning rate before early stop')
    parser.add_argument('--lr-factor', type=float, default=0.8, help='Factor by which to multiply the learning rate when the metric stops improving')
    parser.add_argument('--lr-warmup-steps', type=int, default=0, help='How many steps to warm-up over. Defaults to 0 for no warm-up')
    parser.add_argument('--early-stopping-patience', type=int, default=30, help='Stop training after this many epochs without improvement')
    parser.add_argument('--early-stopping-monitor', type=str, default='val_total_mse_loss', choices=['train_total_mse_loss', 'val_total_mse_loss'], help='Metric to monitor for early stopping')
    parser.add_argument('--reset-trainer', type=bool, default=False, help='Reset training metrics (e.g. early stopping, lr) when loading a model checkpoint')
    parser.add_argument('--weight-decay', type=float, default=0.0, help='Weight decay strength')
    parser.add_argument('--ema-alpha-y', type=float, default=1.0, help='The amount of influence of new losses on the exponential moving average of y. Must be between 0 and 1.')
    parser.add_argument('--ema-alpha-neg-dy', type=float, default=1.0, help='The amount of influence of new losses on the exponential moving average of dy. Must be between 0 and 1.')
    parser.add_argument('--ngpus', type=int, default=-1, help='Number of GPUs, -1 use all available. Use CUDA_VISIBLE_DEVICES=1, to decide gpus')
    parser.add_argument('--num-nodes', type=int, default=1, help='Number of GPU nodes for distributed training with the Lightning Trainer.')
    parser.add_argument('--precision', type=int, default=32, choices=[16, 32, 64], help='Floating point precision')
    parser.add_argument('--log-dir', '-l', default='/tmp/logs', help='log file')
    parser.add_argument('--splits', default=None, help='Npz with splits idx_train, idx_val, idx_test')
    parser.add_argument('--train-size', type=number, default=None, help='Percentage/number of samples in training set (None to use all remaining samples)')
    parser.add_argument('--val-size', type=number, default=0.05, help='Percentage/number of samples in validation set (None to use all remaining samples)')
    parser.add_argument('--test-size', type=number, default=0.1, help='Percentage/number of samples in test set (None to use all remaining samples)')
    parser.add_argument('--test-interval', type=int, default=-1, help='Test interval, one test per n epochs (default: 10)')
    parser.add_argument('--save-interval', type=int, default=10, help='Save interval, one save per n epochs (default: 10)')
    parser.add_argument('--seed', type=int, default=1, help='random seed (default: 1)')
    parser.add_argument('--num-workers', type=int, default=4, help='Number of workers for data prefetch')
    parser.add_argument('--redirect', type=bool, default=False, help='Redirect stdout and stderr to log_dir/log')
    parser.add_argument('--gradient-clipping', type=float, default=0.0, help='Gradient clipping norm')
    parser.add_argument('--remove-ref-energy', action='store_true', help='If true, remove the reference energy from the dataset for delta-learning. Total energy can still be predicted by the model during inference by turning this flag off when loading.  The dataset must be compatible with Atomref for this to be used.')
    parser.add_argument('--checkpoint-monitor', type=str, default='val_total_mse_loss', choices=['train_total_mse_loss', 'val_total_mse_loss'], help='Metric to monitor for writing out best models')
    parser.add_argument('--load-weights', default=None, type=str, help='Load the weights of an existing model')
    # dataset specific
    parser.add_argument('--dataset', default=None, type=str, choices=datasets.__all__, help='Name of the torch_geometric dataset')
    parser.add_argument('--dataset-root', default='~/data', type=str, help='Data storage directory (not used if dataset is "CG")')
    parser.add_argument('--dataset-arg', default=None, help='Additional dataset arguments. Needs to be a dictionary.')
    parser.add_argument('--coord-files', default=None, type=str, help='Custom coordinate files glob')
    parser.add_argument('--embed-files', default=None, type=str, help='Custom embedding files glob')
    parser.add_argument('--energy-files', default=None, type=str, help='Custom energy files glob')
    parser.add_argument('--force-files', default=None, type=str, help='Custom force files glob')
    parser.add_argument('--dataset-preload-limit', default=1024, type=int, help='Custom and HDF5 datasets will preload to RAM datasets that are less than this size in MB')
    parser.add_argument('--y-weight', default=1.0, type=float, help='Weighting factor for y label in the loss function')
    parser.add_argument('--neg-dy-weight', default=1.0, type=float, help='Weighting factor for neg_dy label in the loss function')
    parser.add_argument('--train-loss', default='mse_loss', type=str, choices=loss_class_mapping.keys(), help='Loss function to use during training')
    parser.add_argument('--train-loss-arg', default=None, help='Additional arguments for the loss function. Needs to be a dictionary.')

    # model architecture
    parser.add_argument('--model', type=str, default='graph-network', choices=models.__all_models__, help='Which model to train')
    parser.add_argument('--output-model', type=str, default='Scalar', choices=output_modules.__all__, help='The type of output model')
    parser.add_argument('--output-mlp-num-layers', type=int, default=0, help='If the output model uses an MLP this will be the number of hidden layers, excluding the input and output layers.')
    parser.add_argument('--prior-model', type=str, default=None, help='Which prior model to use. It can be a string, a dict if you want to add arguments for it or a dicts to add more than one prior. e.g. {"Atomref": {"max_z":100}, "Coulomb":{"max_num_neighs"=100, "lower_switch_distance"=4, "upper_switch_distance"=8}', action="extend", nargs="*")

    # architectural args
    parser.add_argument('--charge', type=bool, default=False, help='Model needs a total charge. Set this to True if your dataset contains charges and you want them passed down to the model.')
    parser.add_argument('--spin', type=bool, default=False, help='Model needs a spin state. Set this to True if your dataset contains spin states and you want them passed down to the model.')
    parser.add_argument('--embedding-dimension', type=int, default=256, help='Embedding dimension')
    parser.add_argument('--num-layers', type=int, default=6, help='Number of interaction layers in the model')
    parser.add_argument('--num-rbf', type=int, default=64, help='Number of radial basis functions in model')
    parser.add_argument('--activation', type=str, default='silu', choices=list(act_class_mapping.keys()), help='Activation function')
    parser.add_argument('--rbf-type', type=str, default='expnorm', choices=list(rbf_class_mapping.keys()), help='Type of distance expansion')
    parser.add_argument('--trainable-rbf', type=bool, default=False, help='If distance expansion functions should be trainable')
    parser.add_argument('--neighbor-embedding', type=bool, default=False, help='If a neighbor embedding should be applied before interactions')
    parser.add_argument('--aggr', type=str, default='add', help='Aggregation operation for CFConv filter output. Must be one of \'add\', \'mean\', or \'max\'')

    # Transformer specific
    parser.add_argument('--distance-influence', type=str, default='both', choices=['keys', 'values', 'both', 'none'], help='Where distance information is included inside the attention')
    parser.add_argument('--attn-activation', default='silu', choices=list(act_class_mapping.keys()), help='Attention activation function')
    parser.add_argument('--num-heads', type=int, default=8, help='Number of attention heads')

    # Equivariant Transformer specific
    parser.add_argument('--vector-cutoff', type=bool, default=False, help='If true, the vector features are weighted by the cutoff function during message passing, forcing the energy to be continuous at the cutoff.')

    # TensorNet specific
    parser.add_argument('--equivariance-invariance-group', type=str, default='O(3)', help='Equivariance and invariance group of TensorNet')
    parser.add_argument('--box-vecs', type=lambda x: list(yaml.safe_load(x)), default=None, help="""Box vectors for periodic boundary conditions. The vectors `a`, `b`, and `c` represent a triclinic box and must satisfy
        certain requirements:
        `a[1] = a[2] = b[2] = 0`;`a[0] >= 2*cutoff, b[1] >= 2*cutoff, c[2] >= 2*cutoff`;`a[0] >= 2*b[0]`;`a[0] >= 2*c[0]`;`b[1] >= 2*c[1]`;
        These requirements correspond to a particular rotation of the system and reduced form of the vectors, as well as the requirement that the cutoff be no larger than half the box width.
    Example: [[1,0,0],[0,1,0],[0,0,1]]""")
    parser.add_argument('--static_shapes', type=bool, default=False, help='If true, TensorNet will use statically shaped tensors for the network, making it capturable into a CUDA graphs. In some situations static shapes can lead to a speedup, but it increases memory usage.')

    # other args
    parser.add_argument('--derivative', default=False, type=bool, help='If true, take the derivative of the prediction w.r.t coordinates')
    parser.add_argument('--cutoff-lower', type=float, default=0.0, help='Lower cutoff in model')
    parser.add_argument('--cutoff-upper', type=float, default=5.0, help='Upper cutoff in model')
    parser.add_argument('--atom-filter', type=int, default=-1, help='Only sum over atoms with Z > atom_filter')
    parser.add_argument('--max-z', type=int, default=100, help='Maximum atomic number that fits in the embedding matrix')
    parser.add_argument('--max-num-neighbors', type=int, default=32, help='Maximum number of neighbors to consider in the network')
    parser.add_argument('--standardize', type=bool, default=False, help='If true, multiply prediction by dataset std and add mean')
    parser.add_argument('--reduce-op', type=str, default='add', choices=['add', 'mean'], help='Reduce operation to apply to atomic predictions')
    parser.add_argument('--wandb-use', default=False, type=bool, help='Defines if wandb is used or not')
    parser.add_argument('--wandb-name', default='training', type=str, help='Give a name to your wandb run')
    parser.add_argument('--wandb-project', default='training_', type=str, help='Define what wandb Project to log to')
    parser.add_argument('--wandb-resume-from-id', default=None, type=str, help='Resume a wandb run from a given run id. The id can be retrieved from the wandb dashboard')
    parser.add_argument('--tensorboard-use', default=False, type=bool, help='Defines if tensor board is used or not')

    # fmt: on
    return parser


def get_args():
    parser = get_argparse()
    args = parser.parse_args()
    if args.redirect:
        sys.stdout = open(os.path.join(args.log_dir, "log"), "w")
        sys.stderr = sys.stdout
        logging.getLogger("pytorch_lightning").addHandler(
            logging.StreamHandler(sys.stdout)
        )

    if args.inference_batch_size is None:
        args.inference_batch_size = args.batch_size

    os.makedirs(os.path.abspath(args.log_dir), exist_ok=True)
    save_argparse(args, os.path.join(args.log_dir, "input.yaml"), exclude=["conf"])

    return args


def fix_state_dict(ckpt):
    import re

    state_dict = {re.sub(r"^model\.", "", k): v for k, v in ckpt["state_dict"].items()}
    # In ET, before we had output_model.output_network.{0,1}.update_net.[0-9].{weight,bias}
    # Now we have output_model.output_network.{0,1}.update_net.layers.[0-9].{weight,bias}
    # In other models, we had output_model.output_network.{0,1}.{weight,bias},
    # which is now output_model.output_network.layers.{0,1}.{weight,bias}
    # This change was introduced in https://github.com/torchmd/torchmd-net/pull/314
    patterns = [
        (
            r"output_model.output_network.(\d+).update_net.(\d+).",
            r"output_model.output_network.\1.update_net.layers.\2.",
        ),
        (
            r"output_model.output_network.([02]).(weight|bias)",
            r"output_model.output_network.layers.\1.\2",
        ),
    ]
    for p in patterns:
        state_dict = {re.sub(p[0], p[1], k): v for k, v in state_dict.items()}
    return state_dict


def main():
    args = get_args()
    if args.remove_ref_energy:
        if args.prior_model is None:
            args.prior_model = []
        if not isinstance(args.prior_model, list):
            args.prior_model = [args.prior_model]
        args.prior_model.append({"Atomref": {"enable": False}})

    pl.seed_everything(args.seed, workers=True)

    # initialize data module
    data = DataModule(args)
    data.prepare_data()
    data.setup("fit")

    prior_models = create_prior_models(vars(args), data.dataset)
    args.prior_args = [p.get_init_args() for p in prior_models]
    # initialize lightning module
    model = LNNP(args, prior_model=prior_models, mean=data.mean, std=data.std)
    if args.load_weights is not None:
        print(f"Loading weights from {args.load_weights}")
        ckpt = torch.load(args.load_weights, map_location="cpu")
        model.model.load_state_dict(fix_state_dict(ckpt))

    mon = args.checkpoint_monitor
    chkpoint_name = f"epoch={{epoch}}-{mon.split('_')[0]}_loss={{{mon}:.4f}}"
    if len(data.idx_test):
        chkpoint_name += "-test_loss={test_total_l1_loss:.4f}"

    callbacks = []
    checkpoint_callback = ModelCheckpoint(
        dirpath=args.log_dir,
        monitor=args.checkpoint_monitor,
        save_top_k=10,  # -1 to save all
        every_n_epochs=args.save_interval,
        filename=chkpoint_name,
        auto_insert_metric_name=False,
    )
    callbacks.append(checkpoint_callback)
    if args.early_stopping_monitor is not None:
        early_stopping = EarlyStopping(
            args.early_stopping_monitor, patience=args.early_stopping_patience
        )
        callbacks.append(early_stopping)

    check_logs(args.log_dir)
    csv_logger = CSVLogger(args.log_dir, name="", version="")
    _logger = [csv_logger]

    if args.wandb_use:
        wandb_logger = WandbLogger(
            project=args.wandb_project,
            name=args.wandb_name,
            save_dir=args.log_dir,
            resume="must" if args.wandb_resume_from_id is not None else None,
            id=args.wandb_resume_from_id,
        )
        _logger.append(wandb_logger)

    if args.tensorboard_use:
        tb_logger = TensorBoardLogger(
            args.log_dir, name="tensorboard", version="", default_hp_metric=False
        )
        _logger.append(tb_logger)
    if args.test_interval > 0:
        rank_zero_warn(
            f"WARNING: Test set will be evaluated every {args.test_interval} epochs. This will slow down training."
        )

    trainer = pl.Trainer(
        strategy="auto",
        max_epochs=args.num_epochs,
        accelerator="auto",
        devices=args.ngpus,
        num_nodes=args.num_nodes,
        default_root_dir=args.log_dir,
        callbacks=callbacks,
        logger=_logger,
        precision=args.precision,
        gradient_clip_val=args.gradient_clipping,
        inference_mode=False,
        # Test-during-training requires reloading the dataloaders every epoch
        reload_dataloaders_every_n_epochs=1 if args.test_interval > 0 else 0,
    )

    trainer.fit(model, data, ckpt_path=None if args.reset_trainer else args.load_model)

    # run test set after completing the fit
    model = LNNP.load_from_checkpoint(trainer.checkpoint_callback.best_model_path)
    trainer = pl.Trainer(
        logger=_logger,
        inference_mode=False,
        accelerator="auto",
        devices=args.ngpus,
        num_nodes=args.num_nodes,
    )
    trainer.test(model, data)


if __name__ == "__main__":
    main()
